from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any
import subprocess as sp

from deltona.media import (
    add_info_json_to_media_file,
    archive_dashcam_footage,
    cddb_query,
    create_static_text_video,
    ffprobe,
    get_cd_disc_id,
    get_info_json,
    hlg_to_sdr,
    is_audio_input_format_supported,
    supported_audio_input_formats,
)
import pytest
import requests

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from pytest_mock import MockerFixture
    from requests_mock import Mocker


def test_supported_audio_input_formats_success(mocker: MockerFixture) -> None:
    fake_proc = mocker.Mock()
    fake_proc.stdout = '44100 Hz\n48000 Hz\n'
    fake_proc.stderr = ''
    mocker.patch('deltona.media.sp.run', return_value=fake_proc)
    formats = ('f32le', 's16le')
    rates = (44100, 48000)
    result = supported_audio_input_formats('hw:Audio', formats=formats, rates=rates)
    assert set(result) == {('f32le', 44100), ('f32le', 48000), ('s16le', 44100), ('s16le', 48000)}


def test_supported_audio_input_formats_device_error(mocker: MockerFixture) -> None:
    fake_proc = mocker.Mock()
    fake_proc.stdout = ''
    fake_proc.stderr = 'Device or resource busy'
    mocker.patch('deltona.media.sp.run', return_value=fake_proc)
    with pytest.raises(OSError):  # noqa: PT011
        supported_audio_input_formats('hw:Missing', formats=('f32le',), rates=(44100,))


def test_supported_audio_input_formats_partial_support(mocker: MockerFixture) -> None:
    # Simulate ffmpeg only supporting one format/rate
    def fake_run(cmd: Sequence[str], *args: Any, **kwargs: Any) -> Any:
        if 'pcm_f32le' in cmd and '44100' in cmd:
            return mocker.Mock(stdout='44100 Hz', stderr='')
        return mocker.Mock(stdout='', stderr='cannot set sample format 0x')

    mocker.patch('deltona.media.sp.run', side_effect=fake_run)
    result = supported_audio_input_formats('hw:Audio',
                                           formats=('f32le', 's16le'),
                                           rates=(44100, 48000))
    assert result == (('f32le', 44100),)


def test_supported_audio_input_formats_empty(mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.sp.run')
    result = supported_audio_input_formats('hw:Audio', formats=('f32le',), rates=(44100,))
    assert result == ()


def test_is_audio_input_format_supported_true(mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.supported_audio_input_formats', return_value=(('f32le', 44100),))
    assert is_audio_input_format_supported('hw:Audio', 'f32le', 44100) is True


def test_ffprobe_success(mocker: MockerFixture) -> None:
    fake_proc = mocker.Mock()
    fake_proc.stdout = '{"format": {"tags": {"title": "Test"}}}'
    mocker.patch('deltona.media.sp.run', return_value=fake_proc)
    result = ffprobe('test.flac')
    assert isinstance(result, dict)
    assert result['format']['tags']['title'] == 'Test'  # type: ignore[typeddict-item]


def test_get_info_json_flac(mocker: MockerFixture) -> None:
    # Simulate ffprobe returning info_json in tags
    fake_proc = mocker.Mock()
    fake_proc.stdout = '{"format": {"tags": {"info_json": "{\\"foo\\": \\"bar\\"}"}}}'
    mocker.patch('deltona.media.sp.run', return_value=fake_proc)
    result = get_info_json('test.flac')
    assert isinstance(result, dict)
    assert result['foo'] == 'bar'


def test_get_info_json_mp4(mocker: MockerFixture) -> None:
    # Simulate MP4Box returning info_json string
    fake_proc = mocker.Mock()
    fake_proc.stdout = '{"foo": "bar"}'
    mocker.patch('deltona.media.sp.run', return_value=fake_proc)
    result = get_info_json('test.mp4')
    assert isinstance(result, dict)
    assert result['foo'] == 'bar'


def test_get_info_json_mkv(mocker: MockerFixture) -> None:
    # Simulate mkvextract returning info_json as second line
    fake_proc = mocker.Mock()
    fake_proc.stdout = 'Attachment: info.json\n{"foo": "bar"}'
    mocker.patch('deltona.media.sp.run', return_value=fake_proc)
    result = get_info_json('test.mkv')
    assert isinstance(result, dict)
    assert result['foo'] == 'bar'


def test_get_info_json_mp3(mocker: MockerFixture) -> None:
    # Simulate ffprobe returning TXXX tag with info_json
    fake_proc = mocker.Mock()
    fake_proc.stdout = '{"format": {"tags": {"TXXX": "info_json={\\"foo\\": \\"bar\\"}"}}}'
    mocker.patch('deltona.media.sp.run', return_value=fake_proc)
    result = get_info_json('test.mp3')
    assert isinstance(result, dict)
    assert result['foo'] == 'bar'


def test_get_info_json_opus(mocker: MockerFixture) -> None:
    # Simulate ffprobe returning info_json in streams[0].tags
    fake_proc = mocker.Mock()
    fake_proc.stdout = '{"streams": [{"tags": {"info_json": "{\\"foo\\": \\"bar\\"}"}}]}'
    mocker.patch('deltona.media.sp.run', return_value=fake_proc)
    result = get_info_json('test.opus')
    assert isinstance(result, dict)
    assert result['foo'] == 'bar'


def test_get_info_json_raw_true_flac(mocker: MockerFixture) -> None:
    # Test raw=True returns the raw string
    fake_proc = mocker.Mock()
    fake_proc.stdout = '{"format": {"tags": {"info_json": "{\\"foo\\": \\"bar\\"}"}}}'
    mocker.patch('deltona.media.sp.run', return_value=fake_proc)
    result = get_info_json('test.flac', raw=True)
    assert isinstance(result, str)
    assert result == '{"foo": "bar"}'


def test_get_info_json_not_implemented(mocker: MockerFixture) -> None:
    with pytest.raises(NotImplementedError):
        get_info_json('test.unknown')


def test_create_static_text_video_default_args(mocker: MockerFixture, tmp_path: Path) -> None:
    fake_tempfile = mocker.patch('tempfile.NamedTemporaryFile')
    fake_tf = mocker.Mock()
    fake_tf.name = str(tmp_path / 'temp.png')
    fake_tempfile.return_value.__enter__.return_value = fake_tf
    mock_run = mocker.patch('deltona.media.sp.run')
    mock_unlink = mocker.patch('pathlib.Path.unlink')
    audio_file = tmp_path / 'audio.flac'
    audio_file.write_bytes(b'dummy audio')
    text = 'Test Text'
    create_static_text_video(audio_file, text)
    assert mock_run.call_count >= 1
    mock_unlink.assert_called_with()


def test_create_static_text_video_nvenc(mocker: MockerFixture, tmp_path: Path) -> None:
    fake_tempfile = mocker.patch('tempfile.NamedTemporaryFile')
    fake_tf = mocker.Mock()
    fake_tf.name = str(tmp_path / 'temp.png')
    fake_tempfile.return_value.__enter__.return_value = fake_tf
    mock_run = mocker.patch('deltona.media.sp.run')
    mocker.patch('pathlib.Path.unlink')
    audio_file = tmp_path / 'audio.flac'
    audio_file.write_bytes(b'dummy audio')
    text = 'NVENC Test'
    create_static_text_video(audio_file, text, nvenc=True)
    called_args = [call[0][0] for call in mock_run.call_args_list]
    assert any('h264_nvenc' in str(args) for args in called_args)


def test_create_static_text_video_videotoolbox(mocker: MockerFixture, tmp_path: Path) -> None:
    fake_tempfile = mocker.patch('tempfile.NamedTemporaryFile')
    fake_tf = mocker.Mock()
    fake_tf.name = str(tmp_path / 'temp.png')
    fake_tempfile.return_value.__enter__.return_value = fake_tf

    mock_run = mocker.patch('deltona.media.sp.run')
    mocker.patch('pathlib.Path.unlink')

    audio_file = tmp_path / 'audio.flac'
    audio_file.write_bytes(b'dummy audio')
    text = 'VT Test'

    create_static_text_video(audio_file, text, videotoolbox=True)

    called_args = [call[0][0] for call in mock_run.call_args_list]
    assert any('hevc_videotoolbox' in str(args) for args in called_args)


def test_create_static_text_video_nvenc_and_videotoolbox_error(mocker: MockerFixture,
                                                               tmp_path: Path) -> None:
    audio_file = tmp_path / 'audio.flac'
    audio_file.write_bytes(b'dummy audio')
    text = 'Error Test'
    with pytest.raises(ValueError, match=r'^nvenc and videotoolbox parameters are exclusive.'):
        create_static_text_video(audio_file, text, nvenc=True, videotoolbox=True)


def test_create_static_text_video_sp_run_raises(mocker: MockerFixture, tmp_path: Path) -> None:
    fake_tempfile = mocker.patch('tempfile.NamedTemporaryFile')
    fake_tf = mocker.Mock()
    fake_tf.name = str(tmp_path / 'temp.png')
    fake_tempfile.return_value.__enter__.return_value = fake_tf

    mocker.patch('deltona.media.sp.run', side_effect=[sp.CalledProcessError(1, 'cmd'), None])
    mocker.patch('pathlib.Path.unlink')

    audio_file = tmp_path / 'audio.flac'
    audio_file.write_bytes(b'dummy audio')
    text = 'Fail Test'

    with pytest.raises(sp.CalledProcessError):
        create_static_text_video(audio_file, text)


def test_get_cd_disc_id_linux_success(mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.IS_LINUX', True)  # noqa: FBT003
    fake_fd = 42
    mocker.patch('deltona.media.ctypes.byref')
    mock_context = mocker.patch('deltona.media.context_os_open')
    mock_context.return_value.__enter__.return_value = fake_fd

    fake_libc = mocker.Mock()
    fake_libc.ioctl.side_effect = [0] + [0] * 4  # header + 3 tracks + leadout
    mocker.patch('deltona.media.ctypes.CDLL', return_value=fake_libc)
    header = mocker.patch('deltona.media.CDROMTOCHeader', autospec=True)
    header_instance = header.return_value
    header_instance.cdth_trk1 = 3
    entry = mocker.patch('deltona.media.CDROMTOCEntry', autospec=True)

    def lba_side_effect() -> Iterator[Any]:
        for lba in (100, 200, 300):
            e = mocker.Mock()
            e.cdte_addr.lba = lba
            yield e
        e = mocker.Mock()
        e.cdte_addr.lba = 400
        yield e

    lba_iter = lba_side_effect()
    entry.side_effect = lambda: next(lba_iter)
    mocker.patch('deltona.media.CDROM_LEADOUT', 0xAA)
    mocker.patch('deltona.media.CDROM_LBA', 1)
    mocker.patch('deltona.media.CD_MSF_OFFSET', 150)
    mocker.patch('deltona.media.CD_FRAMES', 75)
    disc_id = get_cd_disc_id('/dev/cdrom')
    assert disc_id == '0d000403 3 250 350 450 7'


def test_get_cd_disc_id_linux_toc_entry_fail_1(mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.IS_LINUX', True)  # noqa: FBT003
    fake_fd = 42
    mocker.patch('deltona.media.ctypes.byref')
    mock_context = mocker.patch('deltona.media.context_os_open')
    mock_context.return_value.__enter__.return_value = fake_fd

    fake_libc = mocker.Mock()
    fake_libc.ioctl.side_effect = [0, -1]
    mocker.patch('deltona.media.ctypes.CDLL', return_value=fake_libc)
    header = mocker.patch('deltona.media.CDROMTOCHeader', autospec=True)
    header_instance = header.return_value
    header_instance.cdth_trk1 = 3
    entry = mocker.patch('deltona.media.CDROMTOCEntry', autospec=True)

    def lba_side_effect() -> Iterator[Any]:
        for lba in (100, 200, 300):
            e = mocker.Mock()
            e.cdte_addr.lba = lba
            yield e
        e = mocker.Mock()
        e.cdte_addr.lba = 400
        yield e

    lba_iter = lba_side_effect()
    entry.side_effect = lambda: next(lba_iter)
    mocker.patch('deltona.media.CDROM_LEADOUT', 0xAA)
    mocker.patch('deltona.media.CDROM_LBA', 1)
    mocker.patch('deltona.media.CD_MSF_OFFSET', 150)
    mocker.patch('deltona.media.CD_FRAMES', 75)
    with pytest.raises(OSError, match=r'^\d+'):
        get_cd_disc_id('/dev/cdrom')


def test_get_cd_disc_id_linux_toc_entry_fail_2(mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.IS_LINUX', True)  # noqa: FBT003
    fake_fd = 42
    mocker.patch('deltona.media.ctypes.byref')
    mock_context = mocker.patch('deltona.media.context_os_open')
    mock_context.return_value.__enter__.return_value = fake_fd

    fake_libc = mocker.Mock()
    fake_libc.ioctl.side_effect = [0, 0, 0, 0, -1]
    mocker.patch('deltona.media.ctypes.CDLL', return_value=fake_libc)
    header = mocker.patch('deltona.media.CDROMTOCHeader', autospec=True)
    header_instance = header.return_value
    header_instance.cdth_trk1 = 3
    entry = mocker.patch('deltona.media.CDROMTOCEntry', autospec=True)

    def lba_side_effect() -> Iterator[Any]:
        for lba in (100, 200, 300):
            e = mocker.Mock()
            e.cdte_addr.lba = lba
            yield e
        e = mocker.Mock()
        e.cdte_addr.lba = 400
        yield e

    lba_iter = lba_side_effect()
    entry.side_effect = lambda: next(lba_iter)
    mocker.patch('deltona.media.CDROM_LEADOUT', 0xAA)
    mocker.patch('deltona.media.CDROM_LBA', 1)
    mocker.patch('deltona.media.CD_MSF_OFFSET', 150)
    mocker.patch('deltona.media.CD_FRAMES', 75)
    with pytest.raises(OSError, match=r'^\d+'):
        get_cd_disc_id('/dev/cdrom')


def test_get_cd_disc_id_not_linux(mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.IS_LINUX', False)  # noqa: FBT003
    with pytest.raises(NotImplementedError):
        get_cd_disc_id('/dev/cdrom')


def test_get_cd_disc_id_ioctl_header_error(mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.ctypes.byref')
    mocker.patch('deltona.media.IS_LINUX', True)  # noqa: FBT003
    mocker.patch('deltona.media.context_os_open').return_value.__enter__.return_value = 42
    fake_libc = mocker.Mock()
    fake_libc.ioctl.side_effect = [-1]
    mocker.patch('deltona.media.ctypes.CDLL', return_value=fake_libc)
    mocker.patch('deltona.media.CDROMTOCHeader', autospec=True)
    with pytest.raises(OSError, match=r'^\d+'):
        get_cd_disc_id('/dev/cdrom')


def test_cddb_query_no_username(mocker: MockerFixture) -> None:
    disc_id = '12345678 1 123 456 789'
    mocker.patch('deltona.media.socket.gethostname', return_value='host')
    mocker.patch('deltona.media.getpass.getuser', return_value=None)
    mocker.patch('deltona.media.keyring.get_password', return_value='host')
    with pytest.raises(ValueError):  # noqa: PT011
        cddb_query(disc_id)


def test_cddb_query_no_host(mocker: MockerFixture) -> None:
    disc_id = '12345678 1 123 456 789'
    mocker.patch('deltona.media.socket.gethostname', return_value='host')
    mocker.patch('deltona.media.getpass.getuser', return_value='username')
    mocker.patch('deltona.media.keyring.get_password', return_value=None)
    with pytest.raises(ValueError):  # noqa: PT011
        cddb_query(disc_id)


def test_cddb_query_success_single_match(mocker: MockerFixture) -> None:
    disc_id = '12345678 2 123 456 789'
    query_response = '200 rock 12345678 Artist / Album / 2020 / 2\n'
    read_response = ("210 Found exact matches, list follows (until terminating `.')\n"
                     'DTITLE=Artist / Album\n'
                     'DYEAR=2020\n'
                     'DGENRE=Rock\n'
                     'TTITLE0=Track One\n'
                     'TTITLE1=Track Two\n'
                     '.\n')
    mocker.patch('deltona.media.socket.gethostname', return_value='host')
    mocker.patch('deltona.media.getpass.getuser', return_value='user')
    mocker.patch('deltona.media.keyring.get_password', return_value='host')

    def fake_requests_get(url: str,
                          params: Any = None,
                          timeout: int | None = None,
                          **kwargs: Any) -> Any:
        class FakeResponse:
            def __init__(self, text: str) -> None:
                self.text = text
                self.status_code = 200
                self.ok = True

            def raise_for_status(self) -> None:
                pass

        if 'cmd=cddb+query' in url or (params and params.get('cmd', '').startswith('cddb query')):
            return FakeResponse(query_response)
        if 'cmd=cddb+read' in url or (params and params.get('cmd', '').startswith('cddb read')):
            return FakeResponse(read_response)
        msg = 'Unexpected URL'
        raise RuntimeError(msg)

    mocker.patch('deltona.media.requests.get', side_effect=fake_requests_get)
    result = cddb_query(disc_id)
    assert result.artist == 'Artist'
    assert result.album == 'Album'
    assert result.year == 2020
    assert result.genre.lower() == 'rock'
    assert result.tracks == ('Track One', 'Track Two')


def test_cddb_query_multiple_matches_accept_first(mocker: MockerFixture,
                                                  requests_mock: Mocker) -> None:
    disc_id = '87654321 3 111 222 333'
    query_response = ("210 Found exact matches, list follows (until terminating `.')\n"
                      'rock 87654321 Artist / Album / 2021 / 3\n'
                      'pop 87654321 Artist2 / Album2 / 2022 / 3\n'
                      '.\n')
    read_response = ("210 Found exact matches, list follows (until terminating `.')\n"
                     'DTITLE=Artist / Album\n'
                     'DYEAR=2021\n'
                     'DGENRE=Rock\n'
                     'TTITLE0=Track A\n'
                     'TTITLE1=Track B\n'
                     'TTITLE2=Track C\n'
                     'OTHER=Some other info\n'
                     '.\n')
    mocker.patch('deltona.media.socket.gethostname', return_value='host')
    mocker.patch('deltona.media.getpass.getuser', return_value='user')
    mocker.patch('deltona.media.keyring.get_password', return_value='host')

    def fake_requests_get(url: str,
                          params: Any = None,
                          timeout: int | None = None,
                          **kwargs: Any) -> Any:
        class FakeResponse:
            def __init__(self, text: str) -> None:
                self.text = text
                self.status_code = 200
                self.ok = True

            def raise_for_status(self) -> None:
                pass

        if 'cmd=cddb+query' in url or (params and params.get('cmd', '').startswith('cddb query')):
            return FakeResponse(query_response)
        if 'cmd=cddb+read' in url or (params and params.get('cmd', '').startswith('cddb read')):
            return FakeResponse(read_response)
        msg = 'Unexpected URL'
        raise RuntimeError(msg)

    mocker.patch('deltona.media.requests.get', side_effect=fake_requests_get)
    result = cddb_query(disc_id, accept_first_match=True)
    assert result.artist == 'Artist'
    assert result.album == 'Album'
    assert result.year == 2021
    assert result.genre.lower() == 'rock'
    assert result.tracks == ('Track A', 'Track B', 'Track C')


def test_cddb_query_multiple_matches_not_accept_first(mocker: MockerFixture,
                                                      requests_mock: Mocker) -> None:
    disc_id = '87654321 3 111 222 333'
    query_response = ("210 Found exact matches, list follows (until terminating `.')\n"
                      'rock 87654321 Artist / Album / 2021 / 3\n'
                      'pop 87654321 Artist2 / Album2 / 2022 / 3\n'
                      '.\n')
    read_response = ("210 Found exact matches, list follows (until terminating `.')\n"
                     'DTITLE=Artist / Album\n'
                     'DYEAR=2021\n'
                     'DGENRE=Rock\n'
                     'TTITLE0=Track A\n'
                     'TTITLE1=Track B\n'
                     'TTITLE2=Track C\n'
                     '.\n')
    mocker.patch('deltona.media.socket.gethostname', return_value='host')
    mocker.patch('deltona.media.getpass.getuser', return_value='user')
    mocker.patch('deltona.media.keyring.get_password', return_value='host')

    def fake_requests_get(url: str,
                          params: Any = None,
                          timeout: int | None = None,
                          **kwargs: Any) -> Any:
        class FakeResponse:
            def __init__(self, text: str) -> None:
                self.text = text
                self.status_code = 200
                self.ok = True

            def raise_for_status(self) -> None:
                pass

        if 'cmd=cddb+query' in url or (params and params.get('cmd', '').startswith('cddb query')):
            return FakeResponse(query_response)
        if 'cmd=cddb+read' in url or (params and params.get('cmd', '').startswith('cddb read')):
            return FakeResponse(read_response)
        msg = 'Unexpected URL'
        raise RuntimeError(msg)

    mocker.patch('deltona.media.requests.get', side_effect=fake_requests_get)
    with pytest.raises(ValueError, match=r'^\d+'):
        cddb_query(disc_id)


def test_cddb_query_no_match_raises(mocker: MockerFixture) -> None:
    disc_id = '00000000 1 0'
    query_response = '202 No match found\n'
    mocker.patch('deltona.media.socket.gethostname', return_value='host')
    mocker.patch('deltona.media.getpass.getuser', return_value='user')
    mocker.patch('deltona.media.keyring.get_password', return_value='host')
    mock_req = mocker.Mock(text=query_response)
    mocker.patch('deltona.media.requests.get', return_value=mock_req)
    with pytest.raises(ValueError, match='202'):
        cddb_query(disc_id)


def test_cddb_query_http_error(mocker: MockerFixture) -> None:
    disc_id = '99999999 1 0'
    mocker.patch('deltona.media.socket.gethostname', return_value='host')
    mocker.patch('deltona.media.getpass.getuser', return_value='user')
    mocker.patch('deltona.media.keyring.get_password', return_value='host')
    mock_get = mocker.patch('deltona.media.requests.get')
    mock_get.return_value.raise_for_status.side_effect = requests.HTTPError
    with pytest.raises(requests.HTTPError):
        cddb_query(disc_id)


def test_hlg_to_sdr_default_args(mocker: MockerFixture, tmp_path: Path) -> None:
    fake_run = mocker.patch('deltona.media.sp.run')
    input_file = tmp_path / 'input.mkv'
    input_file.write_bytes(b'dummy')
    output_file = tmp_path / 'input-sdr.mkv'
    hlg_to_sdr(input_file)
    fake_run.assert_called_once()
    args = fake_run.call_args[0][0]
    assert 'ffmpeg' in args
    assert '-vf' in args
    assert '-c:v' in args
    assert 'libx265' in args
    assert str(output_file) in args


def test_hlg_to_sdr_with_all_args(mocker: MockerFixture, tmp_path: Path) -> None:
    fake_run = mocker.patch('deltona.media.sp.run')
    input_file = tmp_path / 'input.mkv'
    input_file.write_bytes(b'dummy')
    output_file = tmp_path / 'custom.mkv'
    hlg_to_sdr(
        input_file,
        crf=18,
        output_codec='libx264',
        output_file=output_file,
        input_args=['-threads', '2'],
        output_args=['-map', '0'],
        fast=True,
        delete_after=False,
    )
    args = fake_run.call_args[0][0]
    assert 'ffmpeg' in args
    assert '-threads' in args
    assert '-map' in args
    assert '-vf' in args
    assert 'libx264' in args
    assert str(output_file) in args


def test_hlg_to_sdr_delete_after_true(mocker: MockerFixture, tmp_path: Path) -> None:
    mocker.patch('deltona.media.sp.run')
    mocker.patch('deltona.media.send2trash')
    input_file = tmp_path / 'input.mkv'
    input_file.write_bytes(b'dummy')
    hlg_to_sdr(input_file, delete_after=True)


def test_hlg_to_sdr_raises_on_run_error(mocker: MockerFixture, tmp_path: Path) -> None:
    mocker.patch('deltona.media.sp.run', side_effect=sp.CalledProcessError(1, 'ffmpeg'))
    input_file = tmp_path / 'input.mkv'
    input_file.write_bytes(b'dummy')
    with pytest.raises(sp.CalledProcessError):
        hlg_to_sdr(input_file)


def test_archive_dashcam_footage_basic_merge(mocker: MockerFixture, tmp_path: Path) -> None:
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    for i in range(3):
        (front_dir / f'2024051216440{i}_front.mp4').write_bytes(b'front')
        (rear_dir / f'2024051216440{i}_rear.mp4').write_bytes(b'rear')
    mock_run = mocker.patch('deltona.media.sp.run')
    mock_send2trash = mocker.patch('deltona.media.send2trash')
    archive_dashcam_footage(front_dir, rear_dir, output_dir)
    assert mock_run.call_count == 4
    assert mock_send2trash.call_count == 6


def test_archive_dashcam_footage_group_discrepancy_resolution(mocker: MockerFixture,
                                                              tmp_path: Path) -> None:
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    for i in range(2):
        (front_dir / f'2024051216440{i}_front.mp4').write_bytes(b'front')
        (rear_dir / f'2024051216440{i}_rear.mp4').write_bytes(b'rear')
    (rear_dir / '20240612164401_rear.mp4').write_bytes(b'rear')  # extra rear file
    mock_run = mocker.patch('deltona.media.sp.run')
    mock_send2trash = mocker.patch('deltona.media.send2trash')
    archive_dashcam_footage(front_dir,
                            rear_dir,
                            output_dir,
                            allow_group_discrepancy_resolution=True)
    assert mock_run.call_count == 3
    assert mock_send2trash.call_count == 4


def test_archive_dashcam_footage_group_discrepancy_raises(tmp_path: Path) -> None:
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    (front_dir / '20240512164400_front.mp4').write_bytes(b'front')
    (rear_dir / '20240512164400_rear.mp4').write_bytes(b'rear')
    (rear_dir / '20240512164401_rear.mp4').write_bytes(b'rear')
    with pytest.raises(ValueError, match=r'\d+'):
        archive_dashcam_footage(front_dir,
                                rear_dir,
                                output_dir,
                                allow_group_discrepancy_resolution=False)


def test_archive_dashcam_footage_group_discrepancy_raises_2(tmp_path: Path) -> None:
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    (front_dir / '20240512164400_front.mp4').write_bytes(b'front')
    (rear_dir / '20240512164400_rear.mp4').write_bytes(b'rear')
    (rear_dir / '20240515164401_rear.mp4').write_bytes(b'rear')
    with pytest.raises(ValueError, match=r'\d+'):
        archive_dashcam_footage(front_dir,
                                rear_dir,
                                output_dir,
                                allow_group_discrepancy_resolution=False)


def test_archive_dashcam_footage_group_discrepancy_unresolved(tmp_path: Path) -> None:
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    (front_dir / '20240512164400_front.mp4').write_bytes(b'front')
    (rear_dir / '20240512164400_rear.mp4').write_bytes(b'rear')
    (rear_dir / '20240515164401_rear.mp4').write_bytes(b'rear')
    with pytest.raises(ValueError, match=r'\d+'):
        archive_dashcam_footage(front_dir, rear_dir, output_dir)


def test_archive_dashcam_footage_group_discrepancy_solving_bg_len_gt(tmp_path: Path,
                                                                     mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.sp.run')
    mock_trash = mocker.patch('deltona.media.send2trash')
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    last = rear_dir / '20240512164400_rear.mp4'
    (front_dir / '20240512164400_front.mp4').write_bytes(b'front')
    (rear_dir / '20240512164401_rear.mp4').write_bytes(b'rear')
    last.write_bytes(b'rear')
    archive_dashcam_footage(front_dir, rear_dir, output_dir)
    mock_trash.assert_has_calls([mocker.call(last)])


def test_archive_dashcam_footage_group_discrepancy_solving_bg_len_gt_no_delete(
        tmp_path: Path, mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.sp.run')
    mock_trash = mocker.patch('deltona.media.send2trash')
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    last = rear_dir / '20240512164400_rear.mp4'
    (front_dir / '20240512164400_front.mp4').write_bytes(b'front')
    (rear_dir / '20240512164401_rear.mp4').write_bytes(b'rear')
    last.write_bytes(b'rear')
    archive_dashcam_footage(front_dir, rear_dir, output_dir, no_delete=True)
    mock_trash.assert_not_called()


def test_archive_dashcam_footage_group_discrepancy_solving_fg_len_gt(tmp_path: Path,
                                                                     mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.sp.run')
    mock_trash = mocker.patch('deltona.media.send2trash')
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    last = front_dir / '20240512164400_rear.mp4'
    (rear_dir / '20240512164400_front.mp4').write_bytes(b'front')
    (front_dir / '20240512164401_rear.mp4').write_bytes(b'rear')
    (output_dir / '20240512164400_rear.mkv').touch()
    last.write_bytes(b'rear')
    archive_dashcam_footage(front_dir, rear_dir, output_dir)
    mock_trash.assert_has_calls([mocker.call(last)])


def test_archive_dashcam_footage_group_discrepancy_solving_fg_len_gt_no_delete(
        tmp_path: Path, mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.sp.run')
    mock_trash = mocker.patch('deltona.media.send2trash')
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    last = front_dir / '20240512164400_rear.mp4'
    (rear_dir / '20240512164400_front.mp4').write_bytes(b'front')
    (front_dir / '20240512164401_rear.mp4').write_bytes(b'rear')
    (output_dir / '20240512164400_rear.mkv').touch()
    last.write_bytes(b'rear')
    archive_dashcam_footage(front_dir, rear_dir, output_dir, no_delete=True)
    mock_trash.assert_not_called()


def test_archive_dashcam_footage_group_discrepancy_solving_ignores_extra(
        tmp_path: Path, mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.sp.run')
    mock_trash = mocker.patch('deltona.media.send2trash')
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    last = front_dir / '20240512164400_rear.mp4'
    (rear_dir / '20240512164400_front.mp4').write_bytes(b'front')
    (front_dir / '20240512164401_rear.mp4').write_bytes(b'rear')
    (front_dir / '20240512164402_rear.mp4').write_bytes(b'rear')
    last.write_bytes(b'rear')
    archive_dashcam_footage(front_dir, rear_dir, output_dir, overwrite=True)
    assert mock_trash.call_count == 2


def test_archive_dashcam_footage_group_discrepancy_solving_ignores_extra_no_delete(
        tmp_path: Path, mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.sp.run')
    mock_trash = mocker.patch('deltona.media.send2trash')
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    last = front_dir / '20240512164400_rear.mp4'
    (rear_dir / '20240512164400_front.mp4').write_bytes(b'front')
    (front_dir / '20240512164401_rear.mp4').write_bytes(b'rear')
    (front_dir / '20240512164402_rear.mp4').write_bytes(b'rear')
    last.write_bytes(b'rear')
    archive_dashcam_footage(front_dir, rear_dir, output_dir, overwrite=True, no_delete=True)
    assert mock_trash.call_count == 0


def test_archive_dashcam_footage_crash_deletes_unfinished_files(tmp_path: Path,
                                                                mocker: MockerFixture) -> None:
    mocker.patch('deltona.media.sp.run',
                 side_effect=[None, sp.CalledProcessError(1, 'ffmpeg', stderr=b'')])
    mocker.patch('deltona.media.send2trash')
    mock_unlink = mocker.patch('deltona.media.Path.unlink')
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    (rear_dir / '20240512164400_rear.mp4').write_bytes(b'rear')
    (rear_dir / '20240512164401_rear.mp4').write_bytes(b'rear')
    (front_dir / '20240512164400_front.mp4').write_bytes(b'front')
    (front_dir / '20240512164401_front.mp4').write_bytes(b'front')
    with pytest.raises(sp.CalledProcessError):
        archive_dashcam_footage(front_dir, rear_dir, output_dir)
    assert mock_unlink.call_count == 1


def test_archive_dashcam_footage_calls_with_correct_args(mocker: MockerFixture,
                                                         tmp_path: Path) -> None:
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    (front_dir / '0_front.mp4').write_bytes(b'front')
    (rear_dir / '0_rear.mp4').write_bytes(b'rear')
    mock_run = mocker.patch('deltona.media.sp.run')
    mock_send2trash = mocker.patch('deltona.media.send2trash')
    archive_dashcam_footage(front_dir,
                            rear_dir,
                            output_dir,
                            video_encoder='hevc_nvenc',
                            video_bitrate='2M')
    args = mock_run.call_args_list[0].args[0]
    assert 'hevc_nvenc' in args
    assert '-b:v' in args
    assert '2M' in args
    assert mock_send2trash.called


def test_archive_dashcam_footage_calls_with_correct_args_no_delete(mocker: MockerFixture,
                                                                   tmp_path: Path) -> None:
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    (front_dir / '0_front.mp4').write_bytes(b'front')
    (rear_dir / '0_rear.mp4').write_bytes(b'rear')
    mock_run = mocker.patch('deltona.media.sp.run')
    mock_send2trash = mocker.patch('deltona.media.send2trash')
    archive_dashcam_footage(
        front_dir,
        rear_dir,
        output_dir,
        no_delete=True,
        video_encoder='hevc_nvenc',
        video_bitrate='2M',
    )
    args = mock_run.call_args_list[0].args[0]
    assert 'hevc_nvenc' in args
    assert '-b:v' in args
    assert '2M' in args
    assert not mock_send2trash.called


def test_archive_dashcam_footage_skips_hidden_files(mocker: MockerFixture, tmp_path: Path) -> None:
    front_dir = tmp_path / 'front'
    rear_dir = tmp_path / 'rear'
    output_dir = tmp_path / 'output'
    front_dir.mkdir()
    rear_dir.mkdir()
    output_dir.mkdir()
    (front_dir / '.hidden_front.mp4').write_bytes(b'front')
    (rear_dir / '.hidden_rear.mp4').write_bytes(b'rear')
    (front_dir / '0_front.mp4').write_bytes(b'front')
    (rear_dir / '0_rear.mp4').write_bytes(b'rear')
    mocker.patch(
        'deltona.media.group_files',
        side_effect=lambda items, *_: [[Path(x)] for x in sorted(items)],
    )
    mock_run = mocker.patch('deltona.media.sp.run')
    mock_send2trash = mocker.patch('deltona.media.send2trash')
    archive_dashcam_footage(front_dir, rear_dir, output_dir)
    assert mock_run.call_count == 2
    assert mock_send2trash.call_count == 2


@pytest.mark.parametrize('ext', ['flac', 'mp3', 'opus'])
def test_add_info_json_to_media_file_flac_mp3_opus(mocker: MockerFixture, tmp_path: Path,
                                                   ext: str) -> None:
    media_file = tmp_path / f'test.{ext}'
    media_file.write_bytes(b'dummy')
    info_json = tmp_path / 'test.info.json'
    info_json.write_text('{"upload_date": "20220101"}')
    mock_run = mocker.patch('deltona.media.sp.run')
    mocker.patch('deltona.media.copyfile')
    mocker.patch('deltona.media.utime')
    mocker.patch('deltona.media.Path.unlink')
    mocker.patch('deltona.media.Path.exists', return_value=True)
    mocker.patch('deltona.media.Path.open',
                 mocker.mock_open(read_data='{"upload_date": "20220101"}'))
    add_info_json_to_media_file(media_file, info_json)
    assert mock_run.called


def test_add_info_json_to_media_file_mp4(mocker: MockerFixture, tmp_path: Path) -> None:
    media_file = tmp_path / 'test.mp4'
    media_file.write_bytes(b'dummy')
    info_json = tmp_path / 'test.info.json'
    info_json.write_text('{"upload_date": "20220101"}')
    mocker.patch('deltona.media.copyfile')
    mock_run = mocker.patch('deltona.media.sp.run')
    mocker.patch('deltona.media.utime')
    mocker.patch('deltona.media.Path.unlink')
    mocker.patch('deltona.media.Path.exists', return_value=True)
    mocker.patch('deltona.media.Path.open',
                 mocker.mock_open(read_data='{"upload_date": "20220101"}'))
    add_info_json_to_media_file(media_file, info_json)
    assert mock_run.called


def test_add_info_json_to_media_file_mkv_ignores_existing(mocker: MockerFixture,
                                                          tmp_path: Path) -> None:
    media_file = tmp_path / 'test.mkv'
    media_file.write_bytes(b'dummy')
    info_json = tmp_path / 'test.info.json'
    info_json.write_text('{"upload_date": "20220101"}')
    mock_run = mocker.patch('deltona.media.sp.run')
    mock_run.return_value.stdout = (
        "Attachment ID 1: type 'application/json', size 123 bytes, file name 'info.json'")
    mocker.patch('deltona.media.utime')
    mocker.patch('deltona.media.Path.unlink')
    mocker.patch('deltona.media.Path.exists', return_value=True)
    mocker.patch('deltona.media.Path.open',
                 mocker.mock_open(read_data='{"upload_date": "20220101"}'))
    add_info_json_to_media_file(media_file, info_json)
    assert mock_run.called


def test_add_info_json_to_media_file_mkv(mocker: MockerFixture, tmp_path: Path) -> None:
    media_file = tmp_path / 'test.mkv'
    media_file.write_bytes(b'dummy')
    info_json = tmp_path / 'test.info.json'
    info_json.write_text('{"upload_date": "20220101"}')
    mock_run = mocker.patch('deltona.media.sp.run')
    mock_run.return_value.stdout = ''
    mock_utime = mocker.patch('deltona.media.utime')
    mocker.patch('deltona.media.Path.unlink')
    mocker.patch('deltona.media.Path.exists', return_value=True)
    mocker.patch('deltona.media.Path.open',
                 mocker.mock_open(read_data='{"upload_date": "20220101"}'))
    add_info_json_to_media_file(media_file, info_json)
    assert mock_run.call_count == 2
    mock_utime.assert_called_once()


def test_add_info_json_to_media_file_json_path_not_exists(mocker: MockerFixture,
                                                          tmp_path: Path) -> None:
    media_file = tmp_path / 'test.flac'
    mocker.patch('deltona.media.Path.exists', return_value=False)
    mock_log = mocker.patch('deltona.media.log')
    add_info_json_to_media_file(media_file)
    mock_log.warning.assert_called_once()


def test_add_info_json_to_media_file_unknown_extension(mocker: MockerFixture,
                                                       tmp_path: Path) -> None:
    media_file = tmp_path / 'test.unknown'
    info_json = tmp_path / 'test.info.json'
    info_json.write_text('{"upload_date": "20220101"}')
    mocker.patch('deltona.media.Path.exists', return_value=True)
    mocker.patch('deltona.media.Path.open',
                 mocker.mock_open(read_data='{"upload_date": "20220101"}'))
    mock_unlink = mocker.patch('deltona.media.Path.unlink')
    add_info_json_to_media_file(media_file, info_json)
    assert not mock_unlink.called


def test_add_info_json_to_media_file_set_date_handles_missing_upload_date(
        mocker: MockerFixture, tmp_path: Path) -> None:
    media_file = tmp_path / 'test.flac'
    info_json = tmp_path / 'test.info.json'
    info_json.write_text('{}')
    mocker.patch('deltona.media.sp.run')
    mocker.patch('deltona.media.copyfile')
    mock_utime = mocker.patch('deltona.media.utime')
    mocker.patch('deltona.media.Path.unlink')
    mocker.patch('deltona.media.Path.exists', return_value=True)
    mocker.patch('deltona.media.Path.open', mocker.mock_open(read_data='{}'))
    add_info_json_to_media_file(media_file, info_json)
    assert not mock_utime.called


def test_add_info_json_to_media_file_set_date_handles_empty_upload_date(
        mocker: MockerFixture, tmp_path: Path) -> None:
    media_file = tmp_path / 'test.flac'
    info_json = tmp_path / 'test.info.json'
    info_json.write_text('{}')
    mocker.patch('deltona.media.sp.run')
    mocker.patch('deltona.media.copyfile')
    mock_utime = mocker.patch('deltona.media.utime')
    mocker.patch('deltona.media.Path.unlink')
    mocker.patch('deltona.media.Path.exists', return_value=True)
    mocker.patch('deltona.media.Path.open', mocker.mock_open(read_data='{"upload_date": ""}'))
    add_info_json_to_media_file(media_file, info_json)
    assert not mock_utime.called



from __future__ import annotations

from typing import TYPE_CHECKING, Any
import datetime
import os
import subprocess as sp

from deltona.io import (
    RARInfo,
    SFVVerificationError,
    UnRAR,
    UnRARExtractionTestFailed,
    context_os_open,
    extract_gog,
    extract_rar_from_zip,
    make_sfv,
    unpack_0day,
    unpack_ebook,
    verify_sfv,
)
import pytest

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from pytest_mock import MockerFixture

FILE_DESCRIPTOR = 3


def test_context_os_open(mocker: MockerFixture) -> None:
    mock_open = mocker.patch('os.open', return_value=FILE_DESCRIPTOR)
    mock_close = mocker.patch('os.close')
    with context_os_open('test_path', os.O_RDONLY) as fd:
        assert fd == FILE_DESCRIPTOR
        mock_open.assert_called_once_with('test_path', os.O_RDONLY, 511, dir_fd=None)
    mock_close.assert_called_once_with(FILE_DESCRIPTOR)


def test_unpack_0day(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mocker.patch('os.chdir')
    mocker.patch('os.getcwd', return_value='/fake')
    mocker.patch('deltona.io.ZipFile')
    mocker.patch('deltona.io.crc32').return_value = 0
    mock_zip = mocker.Mock()
    mock_diz = mocker.Mock()
    mock_rar = mocker.Mock()
    mock_rar.name = 'test.rar'
    mock_path.return_value.glob.side_effect = [[mock_zip], [mock_diz], [mock_rar], [mock_rar]]
    unpack_0day('test_path')
    mock_diz.unlink.assert_called_once()
    mock_path.return_value.glob.assert_any_call('*.rar')
    mock_path.return_value.glob.assert_any_call('*.[rstuvwxyz][0-9a][0-9r]')
    mock_path.return_value.open.return_value.__enter__.return_value.writelines.assert_any_call(
        mocker.ANY)


def test_unpack_0day_no_diz(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mocker.patch('os.chdir')
    mocker.patch('os.getcwd', return_value='/fake')
    mocker.patch('deltona.io.ZipFile')
    mocker.patch('deltona.io.crc32').return_value = 0
    mock_zip = mocker.Mock()
    mock_rar = mocker.Mock()
    mock_rar.name = 'test.rar'
    mock_path.return_value.glob.side_effect = [[mock_zip], [mock_rar], [mock_rar]]
    unpack_0day('test_path', remove_diz=False)
    mock_path.return_value.glob.assert_any_call('*.rar')
    mock_path.return_value.glob.assert_any_call('*.[rstuvwxyz][0-9a][0-9r]')
    mock_path.return_value.open.return_value.__enter__.return_value.writelines.assert_any_call(
        mocker.ANY)


def test_extract_rar_from_zip_extracts_rar_files(mocker: MockerFixture) -> None:
    mock_zipfile = mocker.Mock()
    # Simulate two rar files and one non-rar file in the zip
    mock_zipfile.namelist.return_value = ['file1.rar', 'file2.r00', 'file3.txt']
    mock_zipfile.extract = mocker.Mock()
    extracted = list(extract_rar_from_zip(mock_zipfile))
    # Should only extract .rar and .r00 files
    assert extracted == ['file1.rar', 'file2.r00']
    mock_zipfile.extract.assert_any_call('file1.rar')
    mock_zipfile.extract.assert_any_call('file2.r00')
    assert mock_zipfile.extract.call_count == 2


def test_extract_rar_from_zip_no_rar_files(mocker: MockerFixture) -> None:
    mock_zipfile = mocker.Mock()
    mock_zipfile.namelist.return_value = ['file1.txt', 'file2.doc']
    mock_zipfile.extract = mocker.Mock()
    extracted = list(extract_rar_from_zip(mock_zipfile))
    assert extracted == []
    mock_zipfile.extract.assert_not_called()


def test_unpack_ebook_success_pdf(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mocker.patch('os.chdir')
    mocker.patch('os.getcwd', return_value='/fake')
    mock_zipfile_cls = mocker.patch('deltona.io.ZipFile')
    mock_extract_rar_from_zip = mocker.patch('deltona.io.extract_rar_from_zip')
    mock_sp_run = mocker.patch('deltona.io.sp.run')
    mocker.patch('deltona.io.log')
    # Setup mocks
    mock_dir = mocker.MagicMock()
    mock_dir.is_dir.return_value = True
    mock_zip1 = mocker.MagicMock()
    mock_zipfile_cls.side_effect = [mock_zip1]
    mock_dir.iterdir.return_value = [mocker.MagicMock(name='file.zip', name__endswith='.zip')]
    mock_rar_file = mocker.MagicMock()
    mock_rar_file.name = 'test.rar'
    mock_extract_rar_from_zip.return_value = ['test.rar']
    # Simulate .pdf file after extraction
    mock_pdf = mocker.MagicMock()
    mock_pdf.name = 'book.pdf'
    mock_pdf.lower = lambda: 'book.pdf'
    mock_pdf.open.return_value.__enter__.return_value.read.return_value = b'%PDF'
    mock_pdf.resolve.return_value.parent.name = 'parent_dir'
    mock_dir.iterdir.side_effect = [
        [mocker.MagicMock(name='file.zip', name__endswith='.zip')],  # for zip_listing
        [],  # for epub_list
        [mock_pdf],  # for pdf_list
    ]
    mock_path.side_effect = [mock_dir, mock_rar_file, mock_pdf]
    # Test
    unpack_ebook('some_path')
    mock_sp_run.assert_called()
    mock_pdf.rename.assert_called_with('../parent_dir.pdf')
    mock_zip1.close.assert_called_once()
    mock_rar_file.unlink.assert_called_once()


def test_unpack_ebook_success_epub(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mocker.patch('os.chdir')
    mocker.patch('os.getcwd', return_value='/fake')
    mock_zipfile_cls = mocker.patch('deltona.io.ZipFile')
    mock_extract_rar_from_zip = mocker.patch('deltona.io.extract_rar_from_zip')
    mock_sp_run = mocker.patch('deltona.io.sp.run')
    mocker.patch('deltona.io.log')
    # Setup mocks
    mock_dir = mocker.Mock()
    mock_dir.is_dir.return_value = True
    mock_zip1 = mocker.Mock()
    mock_zipfile_cls.side_effect = [mock_zip1]
    mock_dir.iterdir.return_value = [mocker.Mock(name='file.zip', name__endswith='.zip')]
    mock_rar_file = mocker.Mock()
    mock_rar_file.name = 'test.rar'
    mock_extract_rar_from_zip.return_value = ['test.rar']
    # Simulate .epub file after extraction
    mock_epub = mocker.Mock()
    mock_epub.name = 'book.epub'
    mock_epub.lower = lambda: 'book.epub'
    mock_epub.resolve.return_value.parent.name = 'parent_dir'
    mock_dir.iterdir.side_effect = [
        [mocker.Mock(name='file.zip', name__endswith='.zip')],  # for zip_listing
        [mock_epub],  # epub_list
        [],  # for pdf_list
    ]
    mock_path.side_effect = [mock_dir, mock_rar_file, mock_epub]
    # Test
    unpack_ebook('some_path')
    mock_sp_run.assert_called()
    mock_epub.rename.assert_called_with('../parent_dir.epub')
    mock_zip1.close.assert_called_once()
    mock_rar_file.unlink.assert_called_once()


def test_unpack_ebook_not_a_directory(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mock_path.return_value.is_dir.return_value = False
    with pytest.raises(NotADirectoryError):
        unpack_ebook('not-a-dir')


def test_unpack_ebook_no_zip_files(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mocker.patch('os.chdir')
    mocker.patch('os.getcwd', return_value='/fake')
    mock_dir = mocker.Mock()
    mock_path.return_value = mock_dir
    mock_dir.is_dir.return_value = True
    mock_dir.iterdir.return_value = []
    with pytest.raises(FileExistsError):
        unpack_ebook('some_path')


def test_unpack_ebook_no_rar_found(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mocker.patch('os.chdir')
    mocker.patch('os.getcwd', return_value='/fake')
    mock_zipfile_cls = mocker.patch('deltona.io.ZipFile')
    mock_extract_rar_from_zip = mocker.patch('deltona.io.extract_rar_from_zip')
    mock_dir = mocker.Mock()
    mock_path.return_value = mock_dir
    mock_dir.is_dir.return_value = True
    mock_zip1 = mocker.Mock()
    mock_zipfile_cls.side_effect = [mock_zip1]
    mock_dir.iterdir.return_value = [mocker.Mock(name='file.zip', name__endswith='.zip')]
    mock_extract_rar_from_zip.return_value = []
    with pytest.raises(ValueError):  # noqa: PT011
        unpack_ebook('some_path')


def test_unpack_ebook_no_pdf_or_epub(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mocker.patch('os.chdir')
    mocker.patch('os.getcwd', return_value='/fake')
    mock_zipfile_cls = mocker.patch('deltona.io.ZipFile')
    mock_extract_rar_from_zip = mocker.patch('deltona.io.extract_rar_from_zip')
    mocker.patch('deltona.io.sp.run')
    mock_dir = mocker.Mock()
    mock_path.return_value = mock_dir
    mock_dir.is_dir.return_value = True
    mock_zip1 = mocker.Mock()
    mock_zipfile_cls.side_effect = [mock_zip1]
    mock_dir.iterdir.return_value = [mocker.Mock(name='file.zip', name__endswith='.zip')]
    mock_rar_file = mocker.Mock()
    mock_rar_file.name = 'test.rar'
    mock_extract_rar_from_zip.return_value = ['test.rar']
    mock_dir.iterdir.side_effect = [
        [mocker.Mock(name='file.zip', name__endswith='.zip')],  # for zip_listing
        [],  # for pdf_list
        [],  # for epub_list
    ]
    with pytest.raises(ValueError):  # noqa: PT011
        unpack_ebook('some_path')


def test_unpack_ebook_more_than_1_pdf(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mocker.patch('os.chdir')
    mocker.patch('os.getcwd', return_value='/fake')
    mock_zipfile_cls = mocker.patch('deltona.io.ZipFile')
    mock_extract_rar_from_zip = mocker.patch('deltona.io.extract_rar_from_zip')
    mocker.patch('deltona.io.sp.run')
    # Setup mocks
    mock_dir = mocker.Mock()
    mock_dir.is_dir.return_value = True
    mock_zip1 = mocker.Mock()
    mock_zipfile_cls.side_effect = [mock_zip1]
    mock_dir.iterdir.return_value = [mocker.Mock(name='file.zip', name__endswith='.zip')]
    mock_rar_file = mocker.Mock()
    mock_rar_file.name = 'test.rar'
    mock_extract_rar_from_zip.return_value = ['test.rar']
    # Simulate multiple .pdf files after extraction
    pdf_files = [
        mocker.Mock(name='book1.pdf', lower=lambda: 'book1.pdf'),
        mocker.Mock(name='book2.pdf', lower=lambda: 'book2.pdf'),
    ]
    for pdf in pdf_files:
        pdf.resolve.return_value.parent.name = 'parent_dir'
    mock_dir.iterdir.side_effect = [
        [mocker.Mock(name='file.zip', name__endswith='.zip')],  # for zip_listing
        [],  # for epub_list
        pdf_files,  # pdf_list with multiple PDFs
    ]
    mock_path.side_effect = [mock_dir, mock_rar_file, *pdf_files]
    # Test
    with pytest.raises(ValueError, match='2'):
        unpack_ebook('some_path')


def test_unpack_ebook_more_than_1_epub(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mocker.patch('os.chdir')
    mocker.patch('os.getcwd', return_value='/fake')
    mock_zipfile_cls = mocker.patch('deltona.io.ZipFile')
    mock_extract_rar_from_zip = mocker.patch('deltona.io.extract_rar_from_zip')
    mocker.patch('deltona.io.sp.run')
    # Setup mocks
    mock_dir = mocker.Mock()
    mock_dir.is_dir.return_value = True
    mock_zip1 = mocker.Mock()
    mock_zipfile_cls.side_effect = [mock_zip1]
    mock_dir.iterdir.return_value = [mocker.Mock(name='file.zip', name__endswith='.zip')]
    mock_rar_file = mocker.Mock()
    mock_rar_file.name = 'test.rar'
    mock_extract_rar_from_zip.return_value = ['test.rar']
    # Simulate multiple .epub files after extraction
    epub_files = [
        mocker.Mock(name='book1.epub', lower=lambda: 'book1.epub'),
        mocker.Mock(name='book2.epub', lower=lambda: 'book2.epub'),
    ]
    for epub in epub_files:
        epub.resolve.return_value.parent.name = 'parent_dir'
    mock_dir.iterdir.side_effect = [
        [mocker.Mock(name='file.zip', name__endswith='.zip')],  # for zip_listing
        epub_files,  # epub_list with multiple EPUBs
        [],  # for pdf_list
    ]
    mock_path.side_effect = [mock_dir, mock_rar_file, *epub_files]
    # Test
    with pytest.raises(ValueError, match='2'):
        unpack_ebook('some_path')


def test_unpack_ebook_pdf_not_pdf(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mocker.patch('os.chdir')
    mocker.patch('os.getcwd', return_value='/fake')
    mock_zipfile_cls = mocker.patch('deltona.io.ZipFile')
    mock_extract_rar_from_zip = mocker.patch('deltona.io.extract_rar_from_zip')
    mocker.patch('deltona.io.sp.run')
    # Setup mocks
    mock_dir = mocker.Mock()
    mock_dir.is_dir.return_value = True
    mock_zip1 = mocker.Mock()
    mock_zipfile_cls.side_effect = [mock_zip1]
    mock_dir.iterdir.return_value = [mocker.Mock(name='file.zip', name__endswith='.zip')]
    mock_rar_file = mocker.Mock()
    mock_rar_file.name = 'test.rar'
    mock_extract_rar_from_zip.return_value = ['test.rar']
    # Simulate non-PDF file after extraction
    mock_pdf = mocker.MagicMock()
    mock_pdf.open.return_value.__enter__.return_value.read.return_value = b'Not a PDF'
    mock_dir.iterdir.side_effect = [
        [mocker.Mock(name='file.zip', name__endswith='.zip')],  # for zip_listing
        [],  # for epub_list
        [mock_pdf],  # pdf_list with non-PDF file
    ]
    mock_path.side_effect = [mock_dir, mock_rar_file, mock_pdf]
    # Test
    with pytest.raises(ValueError, match='Not '):
        unpack_ebook('some_path')


def test_extract_gog_success(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mock_copyfileobj = mocker.patch('deltona.io.shutil.copyfileobj')
    mock_input_path = mocker.MagicMock()
    mock_output_dir = mocker.MagicMock()
    mock_path.side_effect = [mock_output_dir, mock_input_path]
    mock_game_bin = mocker.Mock()
    mock_input_path.resolve.return_value.open.return_value.__enter__.return_value = mock_game_bin
    script = b'#!/bin/sh\noffset=`head -n 5 "$0"`\nfilesizes="1234"\n'
    mock_game_bin.read.side_effect = [
        script,
        script,
        b'mojosetup data',
        b'data zip data',
    ]
    mock_game_bin.seek = mocker.MagicMock()
    mock_game_bin.tell.side_effect = [42]

    def readline_side_effect() -> Iterator[bytes]:
        for _ in range(5):
            yield b'line\n'

    mock_game_bin.readline.side_effect = readline_side_effect()
    mock_unpacker_sh_f = mocker.MagicMock()
    mock_mojosetup_tar_f = mocker.MagicMock()
    mock_datafile_f = mocker.MagicMock()
    mock_output_dir.__truediv__.return_value.open.return_value.__enter__.side_effect = [
        mock_unpacker_sh_f,
        mock_mojosetup_tar_f,
        mock_datafile_f,
    ]
    extract_gog('input.gog', 'output_dir')
    mock_output_dir.mkdir.assert_called_once_with(parents=True)
    assert mock_game_bin.seek.call_count >= 3
    mock_copyfileobj.assert_called_once_with(mock_game_bin, mock_datafile_f)


def test_extract_gog_invalid_offset(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mocker.patch('deltona.io.shutil.copyfileobj')
    mock_input_path = mocker.MagicMock()
    mock_output_dir = mocker.MagicMock()
    mock_path.side_effect = [mock_output_dir, mock_input_path]
    mock_game_bin = mocker.Mock()
    mock_input_path.resolve.return_value.open.return_value.__enter__.return_value = mock_game_bin
    script = b'#!/bin/sh\nfilesizes="1234"\n'
    mock_game_bin.read.return_value = script
    with pytest.raises(ValueError):  # noqa: PT011
        extract_gog('input.gog', 'output_dir')
    mock_output_dir.mkdir.assert_called_once_with(parents=True)


def test_extract_gog_invalid_filesize(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mock_input_path = mocker.MagicMock()
    mock_output_dir = mocker.MagicMock()
    mock_path.side_effect = [mock_output_dir, mock_input_path]
    mock_game_bin = mocker.Mock()
    mock_input_path.resolve.return_value.open.return_value.__enter__.return_value = mock_game_bin
    script = b'#!/bin/sh\noffset=`head -n 5 "$0"`\n'
    mock_game_bin.read.side_effect = [script, script]
    mock_game_bin.seek = mocker.MagicMock()
    mock_game_bin.tell.side_effect = [42]

    def readline_side_effect() -> Iterator[bytes]:
        for _ in range(5):
            yield b'line\n'

    mock_game_bin.readline.side_effect = readline_side_effect()
    mock_unpacker_sh_f = mocker.MagicMock()
    open_f = mock_output_dir.__truediv__.return_value.open.return_value
    open_f.__enter__.return_value = mock_unpacker_sh_f
    with pytest.raises(ValueError):  # noqa: PT011
        extract_gog('input.gog', 'output_dir')
    mock_output_dir.mkdir.assert_called_once_with(parents=True)


def test_verify_sfv_success(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mock_file = mocker.MagicMock()
    mock_path.return_value = mock_file
    # Simulate SFV file with one valid entry
    sfv_content = 'file1.bin ABCDEF12\n'
    mock_file.open.return_value.__enter__.return_value = sfv_content.splitlines(keepends=True)
    mock_file.parent.__truediv__.return_value.read_bytes.return_value = b'data'
    mocker.patch('deltona.io.crc32', return_value=int('ABCDEF12', 16))
    verify_sfv('test.sfv')
    mock_file.open.assert_called_once_with(encoding='utf-8')


def test_verify_sfv_raises_on_crc_mismatch(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mock_file = mocker.MagicMock()
    mock_path.return_value = mock_file
    sfv_content = 'file1.bin 12345678\n'
    mock_file.open.return_value.__enter__.return_value = sfv_content.splitlines(keepends=True)
    mock_file.parent.__truediv__.return_value.read_bytes.return_value = b'wrong'
    mocker.patch('deltona.io.crc32', return_value=int('87654321', 16))
    with pytest.raises(SFVVerificationError) as exc:
        verify_sfv('test.sfv')
    assert 'file1.bin' in str(exc.value)
    assert 'Expected 12345678' in str(exc.value)
    assert 'Actual: 87654321' in str(exc.value)


def test_verify_sfv_ignores_comments_and_blank_lines(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mock_file = mocker.MagicMock()
    mock_path.return_value = mock_file
    sfv_content = ['; This is a comment\n', '# Another comment\n', '\n', 'file2.bin 11111111\n']
    mock_file.open.return_value.__enter__.return_value = sfv_content
    mock_file.parent.__truediv__.return_value.read_bytes.return_value = b'data'
    mocker.patch('deltona.io.crc32', return_value=int('11111111', 16))
    verify_sfv('test.sfv')
    mock_file.open.assert_called_once_with(encoding='utf-8')


def test_verify_sfv_handles_multiple_files(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.io.Path')
    mock_file = mocker.MagicMock()
    mock_path.return_value = mock_file
    sfv_content = ['file1.bin 11111111\n', 'file2.bin 22222222\n']
    mock_file.open.return_value.__enter__.return_value = sfv_content

    def crc32_side_effect(data: bytes) -> int:
        if data == b'data1':
            return int('11111111', 16)
        if data == b'data2':
            return int('22222222', 16)
        return 0

    parent = mock_file.parent
    parent.__truediv__.side_effect = lambda fname: mocker.Mock(read_bytes=mocker.Mock(
        return_value=b'data1' if fname == 'file1.bin' else b'data2'))
    mocker.patch('deltona.io.crc32', side_effect=crc32_side_effect)
    verify_sfv('test.sfv')


def test_make_sfv_writes_header_and_crc_lines(mocker: MockerFixture, tmp_path: Path) -> None:
    mock_file1 = tmp_path / 'file1.bin'
    mock_file1.touch()
    mock_file2 = tmp_path / 'file2.bin'
    mock_file2.touch()
    mock_datetime = mocker.patch('deltona.io.datetime')
    mock_datetime.now.return_value.isoformat.return_value = '2024-06-01 12:00:00'

    def mock_fromtimestamp(ts: str, tz: Any = None) -> Any:
        return mocker.Mock(isoformat=lambda _: f'date{int(ts)}')

    mock_datetime.fromtimestamp.side_effect = mock_fromtimestamp
    mock_crc32 = mocker.patch('deltona.io.crc32')
    mock_crc32.side_effect = [0x11111111, 0x22222222]
    mock_out_file = tmp_path / 'test.sfv'

    make_sfv(mock_out_file, [mock_file1, mock_file2], header=True)
    content = mock_out_file.read_text(encoding='utf-8')
    assert content.startswith('; Generated on')
    assert 'file1.bin 11111111' in content
    assert 'file2.bin 22222222' in content


def test_make_sfv_no_header(mocker: MockerFixture, tmp_path: Path) -> None:
    mock_file = tmp_path / 'file1.bin'
    mock_file.touch()
    mock_out_file = tmp_path / 'test.sfv'
    mocker.patch('deltona.io.crc32', return_value=0xDEADBEEF)
    make_sfv(mock_out_file, [str(mock_file)], header=False)
    written = mock_out_file.read_text(encoding='utf-8')
    assert not written.startswith(';')
    assert 'file1.bin DEADBEEF' in written


def test_make_sfv_empty_files_list(mocker: MockerFixture, tmp_path: Path) -> None:
    mock_file = tmp_path / 'file1.sfv'
    mock_file.touch()
    make_sfv(mock_file, [], header=True)
    content = mock_file.read_text(encoding='utf-8')
    assert content.startswith('; Generated on')


def test_unrar_list_files_parses_output(mocker: MockerFixture) -> None:
    mock_sp_run = mocker.patch('deltona.io.sp.run')
    mock_sp_run.return_value.stdout = """UNRAR 7.11 freeware

Archive: the-archive.rar
Details: RAR 1.5

 Attributes      Size     Date    Time   Name
----------- ---------  ---------- -----  ----
    ..A....    137536  2011-06-07 21:54  xy/file2.bin
    ..A....    381453  2011-06-07 21:54  xy/file1.txt
    ...D...         0  2011-06-07 21:53  xy
----------- ---------  ---------- -----  ----
              8356365                    6
""".strip()
    unrar = UnRAR('unrar')
    files = list(unrar.list_files('archive.rar'))
    assert len(files) == 3
    assert isinstance(files[0], RARInfo)
    assert files[1].name == 'xy/file1.txt'
    assert files[0].name == 'xy/file2.bin'
    assert files[0].size == 137536
    assert files[0].date == datetime.datetime(2011, 6, 7, 21, 54, tzinfo=datetime.timezone.utc)


def test_unrar_test_extraction_success(mocker: MockerFixture) -> None:
    mock_sp_run = mocker.patch('deltona.io.sp.run')
    unrar = UnRAR('unrar')
    # Should not raise
    unrar.test_extraction('archive.rar', 'file.txt')
    mock_sp_run.assert_called_once_with(('unrar', 't', '-y', '-inul', 'archive.rar', 'file.txt'),
                                        check=True)


def test_unrar_test_extraction_failure(mocker: MockerFixture) -> None:
    mocker.patch('deltona.io.sp.run', side_effect=sp.CalledProcessError(1, 'unrar'))
    unrar = UnRAR('unrar')
    with pytest.raises(UnRARExtractionTestFailed):
        unrar.test_extraction('archive.rar', 'file.txt')


def test_unrar_pipe_context_manager(mocker: MockerFixture) -> None:
    mock_popen = mocker.patch('deltona.io.sp.Popen')
    mock_proc = mocker.MagicMock()
    mock_popen.return_value.__enter__.return_value = mock_proc
    unrar = UnRAR('unrar')
    with unrar.pipe('archive.rar', 'file.txt') as proc:
        assert proc is mock_proc
    mock_popen.assert_called_once_with(('unrar', 'p', '-y', '-inul', 'archive.rar', 'file.txt'),
                                       stdout=mocker.ANY,
                                       close_fds=True)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

from deltona.commands.admin import (
    clean_old_kernels_and_modules_main,
    generate_html_dir_tree_main,
    kconfig_to_commands_main,
    kconfig_to_json_main,
    patch_bundle_main,
    reset_tpm_enrollments_main,
    slug_rename_main,
    smv_main,
)
from deltona.system import MultipleKeySlots
from pytest_mock import MockerFixture

if TYPE_CHECKING:
    from click.testing import CliRunner
    from pytest_mock import MockerFixture


def test_reset_tpm_enrollments_main_success(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_reset = mocker.patch('deltona.commands.admin.reset_tpm_enrollment')
    mock_reset.return_value = 0

    result = runner.invoke(reset_tpm_enrollments_main, ['uuid1'])
    assert result.exit_code == 0
    mock_reset.assert_called_once()


def test_reset_tpm_enrollments_main_all(mocker: MockerFixture, runner: CliRunner,
                                        tmp_path: Path) -> None:
    mock_reset = mocker.patch('deltona.commands.admin.reset_tpm_enrollment')
    mock_reset.return_value = 0
    fake_crypttab = tmp_path / 'crypttab'
    fake_crypttab.write_text('# fake-crypttab\n'
                             'name1 UUID=uuid1 /dev/mapper/crypt-root tpm2-device=auto\n'
                             'name2 UUID=uuid2 /dev/mapper/crypt-root tpm2-device=auto\n'
                             '# UUID=uuid3 /dev/mapper/crypt-root tpm2-device=auto')

    result = runner.invoke(reset_tpm_enrollments_main, ['-a', '--crypttab', str(fake_crypttab)])
    assert result.exit_code == 0
    mock_reset.assert_has_calls([
        mocker.call('uuid1', dry_run=True),
        mocker.call('uuid2', dry_run=True),
    ])


def test_reset_tpm_enrollments_main_exception(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_reset = mocker.patch('deltona.commands.admin.reset_tpm_enrollment')
    mock_reset.side_effect = MultipleKeySlots('Unexpected error')

    result = runner.invoke(reset_tpm_enrollments_main, ['uuid1'])
    assert result.exit_code == 0
    assert 'Cannot reset TPM enrolment for' in result.output


def test_clean_old_kernels_and_modules_main_success(mocker: MockerFixture,
                                                    runner: CliRunner) -> None:
    mock_clean = mocker.patch('deltona.commands.admin.clean_old_kernels_and_modules')
    mock_clean.return_value = ['a']

    result = runner.invoke(clean_old_kernels_and_modules_main)
    assert result.exit_code == 0
    mock_clean.assert_called_once()


def test_slug_rename_main_success(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_slug_rename = mocker.patch('deltona.commands.admin.slug_rename')
    mock_slug_rename.return_value = 0

    result = runner.invoke(slug_rename_main, ['old-slug', 'new-slug'])
    assert result.exit_code == 0
    mock_slug_rename.assert_has_calls([
        mocker.call('old-slug', no_lower=False),
        mocker.call('new-slug', no_lower=False),
    ])
    assert 'old-slug ->' not in result.output


def test_slug_rename_main_success_verbose(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_slug_rename = mocker.patch('deltona.commands.admin.slug_rename')
    mock_slug_rename.return_value = 0

    result = runner.invoke(slug_rename_main, ['old-slug', 'new-slug', '-v'])
    assert result.exit_code == 0
    mock_slug_rename.assert_has_calls([
        mocker.call('old-slug', no_lower=False),
        mocker.call('new-slug', no_lower=False),
    ])
    assert 'old-slug ->' in result.output
    assert 'new-slug ->' in result.output


def test_patch_bundle_main_success(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_patch_bundle = mocker.patch('deltona.commands.admin.patch_macos_bundle_info_plist')
    mock_patch_bundle.return_value = 0

    result = runner.invoke(patch_bundle_main, ['bundle-path'])
    assert result.exit_code == 0
    data: dict[str, Any] = {}
    mock_patch_bundle.assert_called_once_with(Path('bundle-path'), **data)


def test_patch_bundle_main_adds_retina(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_patch_bundle = mocker.patch('deltona.commands.admin.patch_macos_bundle_info_plist')
    mock_patch_bundle.return_value = 0

    result = runner.invoke(patch_bundle_main, ['bundle-path', '--retina'])
    assert result.exit_code == 0
    data: dict[str, Any] = {'NSHighResolutionCapable': True}
    mock_patch_bundle.assert_called_once_with(Path('bundle-path'), **data)


def test_patch_bundle_main_adds_env_vars(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_patch_bundle = mocker.patch('deltona.commands.admin.patch_macos_bundle_info_plist')
    mock_patch_bundle.return_value = 0

    result = runner.invoke(patch_bundle_main, ['bundle-path', '-E', 'key1', 'value1'])
    assert result.exit_code == 0
    data: dict[str, Any] = {'LSEnvironment': {'key1': 'value1'}}
    mock_patch_bundle.assert_called_once_with(Path('bundle-path'), **data)


def test_kconfig_to_commands_main_success(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_kconfig_to_commands = mocker.patch('deltona.commands.admin.get_kwriteconfig_commands')
    mock_kconfig_to_commands.return_value = ['a']
    mock_kdeglobals = mocker.MagicMock()
    mock_path = mocker.patch('deltona.commands.admin.Path')
    mock_path.home.return_value.__truediv__.return_value.glob.return_value = []
    mock_path.home.return_value.__truediv__.return_value = mock_kdeglobals
    result = runner.invoke(kconfig_to_commands_main, ['-a'])
    assert result.exit_code == 0
    mock_kconfig_to_commands.assert_called_once_with(mock_kdeglobals)


def test_kconfig_to_commands_main_file_arg(mocker: MockerFixture, runner: CliRunner,
                                           tmp_path: Path) -> None:
    mock_kconfig_to_commands = mocker.patch('deltona.commands.admin.get_kwriteconfig_commands')
    mock_kconfig_to_commands.return_value = ['a']
    mock_file = tmp_path / 'filename'
    mock_file.write_text('content')
    result = runner.invoke(kconfig_to_commands_main, [str(mock_file)])
    assert result.exit_code == 0
    mock_kconfig_to_commands.assert_called_once_with(mock_file)


def test_kconfig_to_json_main_all(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mock_kconfig_to_json = mocker.patch('deltona.commands.admin.get_kconfig_dict')
    mock_kconfig_to_json.return_value = {}
    mock_path = mocker.patch('deltona.commands.admin.Path')
    result = runner.invoke(kconfig_to_json_main, ['--all'])
    assert result.exit_code == 0
    mock_kconfig_to_json.assert_called_with(mock_path.home.return_value.__truediv__.return_value)


def test_kconfig_to_json_main_file_arg(mocker: MockerFixture, runner: CliRunner,
                                       tmp_path: Path) -> None:
    mock_kconfig_to_json = mocker.patch('deltona.commands.admin.get_kconfig_dict')
    mock_kconfig_to_json.return_value = {}
    mock_file = tmp_path / 'filename'
    mock_file.write_text('content')
    result = runner.invoke(kconfig_to_json_main, [str(mock_file)])
    assert result.exit_code == 0
    mock_kconfig_to_json.assert_called_once_with(mock_file)


def test_generate_html_dir_tree_main_success(mocker: MockerFixture, runner: CliRunner,
                                             tmp_path: Path) -> None:
    mock_generate_html_dir_tree = mocker.patch('deltona.commands.admin.generate_html_dir_tree')
    mock_generate_html_dir_tree.return_value = '<html></html>'
    mock_output = tmp_path / 'output.html'
    result = runner.invoke(generate_html_dir_tree_main, [str(tmp_path), '-o', str(mock_output)])
    assert result.exit_code == 0
    mock_generate_html_dir_tree.assert_called_once_with(tmp_path, follow_symlinks=False, depth=2)


def test_smv_main_success(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mock_ssh_client_cls = mocker.patch('deltona.commands.admin._get_ssh_client_cls')
    mock_client = mock_ssh_client_cls.return_value.return_value.__enter__.return_value
    mock_smv = mocker.patch('deltona.commands.admin.secure_move_path')
    mock_smv.return_value = 0
    tmp_src = tmp_path / 'src'
    tmp_src.mkdir()

    result = runner.invoke(smv_main, [str(tmp_src), 'some_host:dst'])
    assert result.exit_code == 0
    mock_smv.assert_called_once_with(mock_client,
                                     tmp_src,
                                     'dst',
                                     dry_run=False,
                                     preserve_stats=False)
    mock_client.load_system_host_keys.assert_called_once()



"""Uncategorised commands."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
import logging
import subprocess as sp

from bascom import setup_logging
from deltona.adp import calculate_salary
from deltona.constants import CONTEXT_SETTINGS
from deltona.io import (
    SFVVerificationError,
    UnRAR,
    UnRARExtractionTestFailed,
    extract_gog,
    unpack_0day,
    verify_sfv,
)
from deltona.typing import INCITS38Code
import click

if TYPE_CHECKING:
    from collections.abc import Sequence

log = logging.getLogger(__name__)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-H',
              '--hours',
              default=160,
              type=int,
              help='Hours worked in a month.',
              metavar='HOURS')
@click.option('-r',
              '--pay-rate',
              default=70.0,
              type=float,
              help='Dollars per hour.',
              metavar='DOLLARS')
@click.option(
    '-s',
    '--state',
    metavar='STATE',
    default='FL',
    type=click.Choice(INCITS38Code.__args__),  # type: ignore[attr-defined]
    help='US state abbreviation.',
)
def adp_main(hours: int = 160,
             pay_rate: float = 70.0,
             state: INCITS38Code = 'FL',
             *,
             debug: bool = False) -> None:
    """Calculate US salary."""
    setup_logging(debug=debug, loggers={'deltona': {}, 'urllib3': {}})
    click.echo(str(calculate_salary(hours=hours, pay_rate=pay_rate, state=state)))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    'dirs',
    nargs=-1,
    metavar='DIR',
    type=click.Path(exists=True, dir_okay=True, file_okay=False, path_type=Path),
)
def unpack_0day_main(dirs: Sequence[Path]) -> None:
    """Unpack RAR files from 0day zip file sets."""
    for path in dirs:
        unpack_0day(path)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('filename', type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option(
    '-o',
    '--output-dir',
    default='.',
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    help='Output directory.',
)
def gogextract_main(filename: Path, output_dir: Path, *, debug: bool = False) -> None:
    """Extract a Linux gog.com archive."""
    setup_logging(debug=debug, loggers={'deltona': {}})
    extract_gog(filename, output_dir)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('rar_filename', type=click.Path(dir_okay=False, exists=True, path_type=Path))
@click.option('--no-crc-check', is_flag=True, help='Disable CRC check.')
@click.option('--test-extraction', help='Enable extraction test.', is_flag=True)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option(
    '-D',
    '--device-name',
    help='Device name.',
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option('-s', '--speed', type=int, help='Disc write speed.', default=8)
@click.option('--sfv',
              help='SFV file.',
              type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.option('--cdrecord-path', help='Path to cdrecord.', default='cdrecord')
@click.option('--unrar-path', help='Path to unrar.', default='unrar')
def burnrariso_main(
    rar_filename: Path,
    unrar_path: str = 'unrar',
    cdrecord_path: str = 'cdrecord',
    device_name: Path | None = None,
    sfv: Path | None = None,
    speed: int = 8,
    *,
    debug: bool = False,
    no_crc_check: bool = False,
    test_extraction: bool = False,
) -> None:
    """Burns an ISO found in a RAR file via piping."""  # noqa: DOC501
    setup_logging(debug=debug, loggers={'deltona': {}})
    rar_path = Path(rar_filename)
    unrar = UnRAR(unrar_path)
    isos = [x for x in unrar.list_files(rar_path) if x.name.lower().endswith('.iso')]
    if len(isos) != 1:
        raise click.Abort
    iso = isos[0]
    if not iso.size:
        raise click.Abort
    if not no_crc_check:
        sfv_file_expected = (Path(sfv) if sfv else rar_path.parent /
                             f'{rar_path.name.split(".", 1)}.sfv')
        assert sfv_file_expected.exists()
        try:
            verify_sfv(sfv_file_expected)
        except SFVVerificationError as e:
            click.echo('SFV verification failed.', err=True)
            raise click.Abort from e
    if test_extraction:
        click.echo('Testing extraction.')
        try:
            unrar.test_extraction(rar_path, iso.name)
        except UnRARExtractionTestFailed as e:
            click.echo('RAR extraction test failed.', err=True)
            raise click.Abort from e
    with (
            unrar.pipe(rar_filename, iso.name) as u,
            sp.Popen(
                (
                    cdrecord_path,
                    *((f'dev={device_name}',) if device_name else ()),
                    f'speed={speed}',
                    'driveropts=burnfree',
                    f'tsize={iso.size}',
                ),
                stdin=u.stdout,
                close_fds=True,
            ) as cdrecord,
    ):
        assert u.stdout is not None
        u.stdout.close()
        cdrecord.wait()
        u.wait()
        if not (u.returncode == 0 and cdrecord.returncode == 0):
            click.echo('Write failed!', err=True)
            raise click.Abort



from __future__ import annotations

from http import HTTPStatus
from typing import TYPE_CHECKING, Any
import plistlib

from deltona.www import (
    KEY_ORIGIN_URL,
    KEY_WHERE_FROMS,
    BookmarksDataset,
    BookmarksHTMLFolderAttributes,
    check_bookmarks_html_urls,
    create_parsed_tree_structure,
    generate_html_dir_tree,
    parse_bookmarks_html,
    upload_to_imgbb,
    where_from,
)
from requests import HTTPError
import pytest

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from bs4 import BeautifulSoup
    from pytest_mock import MockerFixture
    from requests_mock import Mocker


def test_upload_to_imgbb_with_api_key(tmp_path: Path, mocker: MockerFixture,
                                      requests_mock: Mocker) -> None:
    # Create a dummy image file
    img_path = tmp_path / 'test.png'
    img_path.write_bytes(b'fake-image-data')

    # Mock requests.post
    url = 'https://api.imgbb.com/1/upload'
    expected_response = {'data': {'url': 'https://imgbb.com/test.png'}}
    requests_mock.post(url, json=expected_response, status_code=200)

    # Call function
    r = upload_to_imgbb(str(img_path), api_key='dummy-key', timeout=1)
    assert r.status_code == 200
    assert r.json() == expected_response


def test_upload_to_imgbb_with_keyring(tmp_path: Path, mocker: MockerFixture,
                                      requests_mock: Mocker) -> None:
    # Create a dummy image file
    img_path = tmp_path / 'test2.png'
    img_path.write_bytes(b'another-fake-image-data')

    # Mock keyring.get_password
    mocker.patch('keyring.get_password', return_value='keyring-key')
    # Mock getuser
    mocker.patch('deltona.www.getuser', return_value='testuser')

    # Mock requests.post
    url = 'https://api.imgbb.com/1/upload'
    expected_response = {'data': {'url': 'https://imgbb.com/test2.png'}}
    requests_mock.post(url, json=expected_response, status_code=200)

    # Call function without api_key, but with keyring_username
    r = upload_to_imgbb(str(img_path), keyring_username='testuser', timeout=1)
    assert r.status_code == 200
    assert r.json() == expected_response


def test_upload_to_imgbb_raises_for_status(tmp_path: Path, mocker: MockerFixture,
                                           requests_mock: Mocker) -> None:
    img_path = tmp_path / 'fail.png'
    img_path.write_bytes(b'fail-data')

    requests_mock.post('https://api.imgbb.com/1/upload',
                       status_code=400,
                       json={'error': 'bad request'})
    with pytest.raises(HTTPError):
        upload_to_imgbb(str(img_path), api_key='bad-key', timeout=1)


def test_check_bookmarks_html_urls_basic(mocker: MockerFixture, requests_mock: Mocker) -> None:
    # Minimal bookmarks.html with two links: one valid, one 404
    html = """
    <DL>
        <DT><A HREF="https://example.com" ADD_DATE="123">Example</A>
        <DT><A HREF="https://notfound.com" ADD_DATE="456">NotFound</A>
    </DL>
    """
    # Mock HEAD requests
    requests_mock.head('https://example.com', status_code=200)
    requests_mock.head('https://notfound.com', status_code=404)

    # Patch user agent generator to avoid dependency
    mocker.patch('deltona.www.generate_chrome_user_agent', return_value='UA')

    data, changed, not_found = check_bookmarks_html_urls(html)
    # Should have both links in data
    assert len(data) == 2
    # No changed links
    assert changed == []
    # One not found
    assert len(not_found) == 1
    assert not_found[0]['attrs']['href'] == 'https://notfound.com'  # type: ignore[typeddict-item]
    assert not_found[0]['title'] == 'NotFound'  # type: ignore[typeddict-item]


def test_check_bookmarks_html_urls_redirect(mocker: MockerFixture, requests_mock: Mocker) -> None:
    html = """
    <DL>
        <DT><A HREF="https://redirect.com" ADD_DATE="789">Redirected</A>
        <DT><A>Invalid</A>
    </DL>
    """
    requests_mock.head('https://redirect.com', status_code=301, headers={'location': '/new-loc'})
    # Patch user agent generator
    mocker.patch('deltona.www.generate_chrome_user_agent', return_value='UA')

    data, changed, not_found = check_bookmarks_html_urls(html)
    # Should have one link in data
    assert len(data) == 2
    # Should have one changed link
    assert len(changed) == 1
    # The href should be rewritten to absolute
    assert changed[0]['attrs']['href'].endswith('/new-loc')  # type: ignore[typeddict-item]
    assert changed[0]['title'] == 'Redirected'  # type: ignore[typeddict-item]
    # No not found
    assert not_found == []


def test_check_bookmarks_html_urls_full_redirect(mocker: MockerFixture,
                                                 requests_mock: Mocker) -> None:
    html = """
    <DL>
        <DT><A HREF="https://redirect.com" ADD_DATE="789">Redirected</A>
    </DL>
    """
    requests_mock.head('https://redirect.com',
                       status_code=301,
                       headers={'location': 'https://new-host/new-loc'})
    mocker.patch('deltona.www.generate_chrome_user_agent', return_value='UA')
    data, changed, not_found = check_bookmarks_html_urls(html)
    assert len(data) == 1
    assert len(changed) == 1
    assert changed[0]['attrs']['href'].endswith('/new-loc')  # type: ignore[typeddict-item]
    assert changed[0]['title'] == 'Redirected'  # type: ignore[typeddict-item]
    assert not_found == []


def test_check_bookmarks_html_urls_exhaustive_check(mocker: MockerFixture) -> None:
    html = """
<!DOCTYPE NETSCAPE-Bookmark-file-1>
<!-- This is an automatically generated file.
     It will be read and overwritten.
     DO NOT EDIT! -->
    <META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=UTF-8">
    <TITLE>Bookmarks</TITLE>
    <H1>Bookmarks</H1>
    <DL><p>
        <DT><H3 ADD_DATE="1649305257" LAST_MODIFIED="1741660172" PERSONAL_TOOLBAR_FOLDER="true">
            Bookmarks bar</H3>
        <DT><A HREF="https://mail.google.com/"></A>
        <DT><A HREF="https://github.com/issues"></A>
    <DL><p>
        <DT><A HREF="https://forums.mydigitallife.net">
        <DT><A HREF="https://deltona.dev">Deltona</A>
    <DL><p>
        <DT><A HREF="https://deltona.zzz">Deltona</A>
        <DT><A HREF="https://deltona.yyy/docs">Documentation</A>
        <DT><H3 ADD_DATE="1620763254" LAST_MODIFIED="1697216205">Other folder</H3>
        <DL><p>
            <DT><A HREF="https://deltona.fff/docs/installation">Installation</A>
            <DT><A HREF="https://deltona.ggg/docs/usage">Usage</A>
        </DL><p>
    </DL><p>
    <DT><A HREF="https://deltona.dev/blog">Blog</A>
    <DT><A HREF="https://deltona.dev/contact">Contact</A>
</DL><p>
<DT><H3 ADD_DATE="1620763254" LAST_MODIFIED="1697216205">Downloads</H3>
<DL><p>
    <DT><A HREF="https://deltona.dev/downloads">Deltona Downloads</A>
    <DT><A HREF="https://deltona.dev/downloads/deltona-0.1.0.tar.gz">Deltona 0.1.0</A>
    <DT><A HREF="https://deltona.dev/downloads/deltona-0.2.0.tar.gz">Deltona 0.2.0</A>
</DL><p>
    """
    mocker.patch('deltona.www.generate_chrome_user_agent', return_value='UA')
    mock_session = mocker.patch('deltona.www.requests.Session')
    n = 0

    def mock_head(url: str, **kwargs: Any) -> Any:
        nonlocal n
        if url.startswith('https://deltona.dev'):
            return mocker.MagicMock(status_code=HTTPStatus.OK)
        n += 1
        return mocker.MagicMock(
            status_code=HTTPStatus.FOUND if n % 3 == 0 else HTTPStatus.NOT_FOUND if n %
            2 == 0 else HTTPStatus.OK,
            headers={'location': 'https://deltona.dev' if n % 2 == 0 else '/index.html'},
        )

    mock_session.return_value.head.side_effect = mock_head
    data, changed, not_found = check_bookmarks_html_urls(html)
    mock_session.return_value.head.assert_has_calls([
        mocker.call('https://mail.google.com/'),
        mocker.call('https://github.com/issues'),
        mocker.call('https://forums.mydigitallife.net'),
        mocker.call('https://deltona.dev'),
        mocker.call('https://deltona.zzz'),
        mocker.call('https://deltona.yyy/docs'),
        mocker.call('https://deltona.fff/docs/installation'),
        mocker.call('https://deltona.ggg/docs/usage'),
        mocker.call('https://deltona.dev/blog'),
        mocker.call('https://deltona.dev/contact'),
        mocker.call('https://deltona.dev/downloads'),
        mocker.call('https://deltona.dev/downloads/deltona-0.1.0.tar.gz'),
        mocker.call('https://deltona.dev/downloads/deltona-0.2.0.tar.gz'),
    ])
    assert data == [
        {
            'type': 'link',
            'title': '',
            'attrs': {
                'href': 'https://mail.google.com/'
            }
        },
        {
            'type': 'link',
            'title': '',
            'attrs': {
                'href': 'https://github.com/issues'
            }
        },
        {
            'type': 'link',
            'title': '',
            'attrs': {
                'href': 'https://forums.mydigitallife.net/index.html'
            },
        },
        {
            'type': 'link',
            'title': 'Deltona',
            'attrs': {
                'href': 'https://deltona.dev'
            }
        },
        {
            'type': 'link',
            'title': 'Deltona',
            'attrs': {
                'href': 'https://deltona.zzz'
            }
        },
        {
            'type': 'link',
            'title': 'Documentation',
            'attrs': {
                'href': 'https://deltona.yyy/docs'
            }
        },
        {
            'attrs': {
                'add_date': '1620763254',
                'last_modified': '1697216205'
            },
            'children': [{
                'type': 'link',
                'title': 'Installation',
                'attrs': {
                    'href': 'https://deltona.dev'
                }
            }],
            'name': 'Other folder',
            'type': 'folder',
        },
        {
            'type': 'link',
            'title': 'Usage',
            'attrs': {
                'href': 'https://deltona.ggg/docs/usage'
            }
        },
        {
            'type': 'link',
            'title': 'Blog',
            'attrs': {
                'href': 'https://deltona.dev/blog'
            }
        },
        {
            'type': 'link',
            'title': 'Contact',
            'attrs': {
                'href': 'https://deltona.dev/contact'
            }
        },
        {
            'attrs': {
                'add_date': '1620763254',
                'last_modified': '1697216205'
            },
            'children': [{
                'type': 'link',
                'title': 'Deltona Downloads',
                'attrs': {
                    'href': 'https://deltona.dev/downloads'
                },
            }],
            'name': 'Downloads',
            'type': 'folder',
        },
        {
            'type': 'link',
            'title': 'Deltona 0.1.0',
            'attrs': {
                'href': 'https://deltona.dev/downloads/deltona-0.1.0.tar.gz'
            },
        },
        {
            'type': 'link',
            'title': 'Deltona 0.2.0',
            'attrs': {
                'href': 'https://deltona.dev/downloads/deltona-0.2.0.tar.gz'
            },
        },
    ]
    assert len(changed) == 2
    assert len(not_found) == 2


def test_where_from_linux(mocker: MockerFixture) -> None:
    # Simulate IS_LINUX = True
    mocker.patch('deltona.www.IS_LINUX', True)  # noqa: FBT003
    # Mock getxattr to return bytes
    mock_getxattr = mocker.patch('deltona.www._getxattr', return_value=b'https://example.com')
    # Should return the decoded string
    result = where_from('dummy-file')
    assert result == 'https://example.com'
    mock_getxattr.assert_called_once_with('dummy-file', KEY_ORIGIN_URL)


def test_where_from_macos_webpage_false(mocker: MockerFixture) -> None:
    # Simulate IS_LINUX = False
    mocker.patch('deltona.www.IS_LINUX', False)  # noqa: FBT003
    # Prepare fake plist data
    fake_plist = plistlib.dumps(['https://file.com', 'https://webpage.com'])
    # Patch hexstr2bytes to just return the bytes
    mocker.patch('deltona.www.hexstr2bytes', return_value=fake_plist)
    # Mock getxattr to return a dummy value (will be passed to hexstr2bytes)
    mock_getxattr = mocker.patch('deltona.www._getxattr', return_value=b'dummy')
    # Should return the first item (index 0)
    result = where_from('dummy-file', webpage=False)
    assert result == 'https://file.com'
    mock_getxattr.assert_called_once_with('dummy-file', KEY_WHERE_FROMS)


def test_where_from_macos_webpage_true(mocker: MockerFixture) -> None:
    # Simulate IS_LINUX = False
    mocker.patch('deltona.www.IS_LINUX', False)  # noqa: FBT003
    # Prepare fake plist data
    fake_plist = plistlib.dumps(['https://file.com', 'https://webpage.com'])
    # Patch hexstr2bytes to just return the bytes
    mocker.patch('deltona.www.hexstr2bytes', return_value=fake_plist)
    # Mock getxattr to return a dummy value (will be passed to hexstr2bytes)
    mock_getxattr = mocker.patch('deltona.www._getxattr', return_value=b'dummy')
    # Should return the second item (index 1)
    result = where_from('dummy-file', webpage=True)
    assert result == 'https://webpage.com'
    mock_getxattr.assert_called_once_with('dummy-file', KEY_WHERE_FROMS)


def test_generate_html_dir_tree_basic(tmp_path: Path) -> None:
    # Create a simple directory structure
    d1 = tmp_path / 'dir1'
    d1.mkdir()
    f1 = d1 / 'file1.txt'
    f1.write_text('hello')
    f2 = tmp_path / 'file2.txt'
    f2.write_text('world')

    # Patch Path.iterdir to avoid OS-specific ordering
    # (but here we use the real filesystem)

    html = generate_html_dir_tree(tmp_path)
    # Should contain both file2.txt and dir1/file1.txt
    assert 'file2.txt' in html
    assert 'file1.txt' in html
    assert 'dir1' in html
    # Should have <ul> and <li> tags
    assert '<ul' in html
    assert '<li' in html


def test_generate_html_dir_tree_nested(tmp_path: Path) -> None:
    # Create nested directories
    d1 = tmp_path / 'a'
    d1.mkdir()
    d2 = d1 / 'b'
    d2.mkdir()
    f1 = d2 / 'c.txt'
    f1.write_text('nested')

    html = generate_html_dir_tree(tmp_path)
    # Should contain all directory and file names
    assert 'a' in html
    assert 'b' in html
    assert 'c.txt' in html


def test_generate_html_dir_tree_empty_dir(tmp_path: Path) -> None:
    # Empty directory

    html = generate_html_dir_tree(tmp_path)
    # Should still return valid HTML
    assert '<ul' in html
    assert '</ul>' in html


def test_generate_html_dir_tree_symlink(tmp_path: Path) -> None:
    # Create a file and a symlink to it
    f1 = tmp_path / 'real.txt'
    f1.write_text('data')
    symlink = tmp_path / 'link.txt'
    symlink.symlink_to(f1)

    html = generate_html_dir_tree(tmp_path)
    # Should include both the real file and the symlink
    assert 'real.txt' in html
    assert 'link.txt' in html


def test_parse_bookmarks_html(mocker: MockerFixture) -> None:
    def recurse_bookmarks_html(soup: BeautifulSoup, callback: Callable[..., Any]) -> None:
        callback(mocker.MagicMock(), 'title', ['a', 'b'])

    mocker.patch('deltona.www.create_parsed_tree_structure', return_value=[])
    mocker.patch('deltona.www.recurse_bookmarks_html', recurse_bookmarks_html)
    ret = parse_bookmarks_html('')
    assert ret == []


def test_create_parsed_tree_structure_creates_new_folders(mocker: MockerFixture) -> None:
    folder_path: list[tuple[str, BookmarksHTMLFolderAttributes]] = [
        ('Folder1', {
            'add_date': '1',
            'last_modified': '2'
        }),
        ('Folder2', {
            'add_date': '3',
            'last_modified': '4'
        }),
    ]
    data: BookmarksDataset = []
    result = create_parsed_tree_structure(folder_path, data)
    assert isinstance(result, list)
    assert len(data) == 1
    assert data[0]['name'] == 'Folder1'  # type: ignore[typeddict-item]
    assert data[0]['type'] == 'folder'
    assert 'children' in data[0]
    assert data[0]['children'][0]['name'] == 'Folder2'  # type: ignore[typeddict-item]
    assert data[0]['children'][0]['type'] == 'folder'
    assert result is data[0]['children'][0]['children']



"""Main CLI group wrapping all commands."""

from __future__ import annotations

from importlib import import_module
from typing import Any
import sys

from deltona.constants import CONTEXT_SETTINGS
import click

_COMMANDS: dict[str, str] = {
    'add-cdda-times': 'deltona.commands.media:add_cdda_times_main',
    'add-info-json': 'deltona.commands.media:add_info_json_main',
    'adp': 'deltona.commands.misc:adp_main',
    'audio2vid': 'deltona.commands.media:audio2vid_main',
    'burnrariso': 'deltona.commands.misc:burnrariso_main',
    'cddb-query': 'deltona.commands.media:cddb_query_main',
    'check-bookmarks-html': 'deltona.commands.www:check_bookmarks_html_main',
    'chrome-bisect-flags': 'deltona.commands.www:chrome_bisect_flags_main',
    'clean-old-kernels-modules': 'deltona.commands.admin:clean_old_kernels_and_modules_main',
    'connect-g603': 'deltona.commands.desktop:connect_g603_main',
    'cssq': 'deltona.commands.string:cssq_main',
    'display-info-json': 'deltona.commands.media:display_info_json_main',
    'encode-dashcam': 'deltona.commands.media:encode_dashcam_main',
    'fix-pwa-icon': 'deltona.commands.www:fix_chromium_pwa_icon_main',
    'flac-dir-finalize': 'deltona.commands.media:flac_dir_finalize_main',
    'fullwidth2ascii': 'deltona.commands.string:fullwidth2ascii_main',
    'git-checkout-default-branch': 'deltona.commands.git:git_checkout_default_branch_main',
    'git-co-default-branch': 'deltona.commands.git:git_checkout_default_branch_main',
    'git-open': 'deltona.commands.git:git_open_main',
    'git-rebase-default-branch': 'deltona.commands.git:git_rebase_default_branch_main',
    'gogextract': 'deltona.commands.misc:gogextract_main',
    'hlg2sdr': 'deltona.commands.media:hlg2sdr_main',
    'htmltree': 'deltona.commands.admin:generate_html_dir_tree_main',
    'inhibit-notifications': 'deltona.commands.desktop:inhibit_notifications_main',
    'is-ascii': 'deltona.commands.string:is_ascii_main',
    'is-bin': 'deltona.commands.string:is_bin_main',
    'json2yaml': 'deltona.commands.string:json2yaml_main',
    'kconfig-to-commands': 'deltona.commands.admin:kconfig_to_commands_main',
    'kconfig-to-json': 'deltona.commands.admin:kconfig_to_json_main',
    'ke-ebook-ex': 'deltona.commands.media:ke_ebook_ex_main',
    'kill-gamescope': 'deltona.commands.desktop:kill_gamescope_main',
    'kill-wine': 'deltona.commands.wine:kill_wine_main',
    'merge-dependabot-prs': 'deltona.commands.git:merge_dependabot_prs_main',
    'mp4json': 'deltona.commands.media:add_info_json_main',
    'mp4json-display': 'deltona.commands.media:display_info_json_main',
    'mpv-sbs': 'deltona.commands.desktop:mpv_sbs_main',
    'mvid-rename': 'deltona.commands.media:mvid_rename_main',
    'netloc': 'deltona.commands.string:urldecode_main',
    'patch-bundle': 'deltona.commands.admin:patch_bundle_main',
    'patch-uiso-font': 'deltona.commands.wine:patch_ultraiso_font_main',
    'pl2json': 'deltona.commands.string:pl2json_main',
    'sanitize': 'deltona.commands.string:sanitize_main',
    'set-wine-fonts': 'deltona.commands.wine:set_wine_fonts_main',
    'show-info-json': 'deltona.commands.media:display_info_json_main',
    'slug-rename': 'deltona.commands.admin:slug_rename_main',
    'slugify': 'deltona.commands.string:slugify_main',
    'smv': 'deltona.commands.admin:smv_main',
    'supported-audio-input-formats': 'deltona.commands.media:supported_audio_input_formats_main',
    'systemd-reset-tpm-cryptenroll': 'deltona.commands.admin:reset_tpm_enrollments_main',
    'tbc2srt': 'deltona.commands.media:tbc2srt_main',
    'title-fixer': 'deltona.commands.string:title_fixer_main',
    'trim': 'deltona.commands.string:trim_main',
    'ucwords': 'deltona.commands.string:ucwords_main',
    'uiso': 'deltona.commands.media:ultraiso_main',
    'umpv': 'deltona.commands.desktop:umpv_main',
    'underscorize': 'deltona.commands.string:underscorize_main',
    'unix2wine': 'deltona.commands.wine:unix2wine_main',
    'unpack-0day': 'deltona.commands.misc:unpack_0day_main',
    'unregister-wine-assocs': 'deltona.commands.wine:unregister_wine_file_associations_main',
    'upload-to-imgbb': 'deltona.commands.desktop:upload_to_imgbb_main',
    'urldecode': 'deltona.commands.string:urldecode_main',
    'wait-for-disc': 'deltona.commands.media:wait_for_disc_main',
    'where-from': 'deltona.commands.www:where_from_main',
    'winegoginstall': 'deltona.commands.wine:winegoginstall_main',
    'wineshell': 'deltona.commands.wine:wineshell_main',
}
_LINUX_ONLY: frozenset[str] = frozenset({
    'clean-old-kernels-modules',
    'connect-g603',
    'inhibit-notifications',
    'kill-gamescope',
    'systemd-reset-tpm-cryptenroll',
    'wait-for-disc',
})
_NOT_WINDOWS: frozenset[str] = _LINUX_ONLY | frozenset({
    'kill-wine',
    'set-wine-fonts',
    'unregister-wine-assocs',
    'winegoginstall',
    'wineshell',
})


def _excluded() -> frozenset[str]:
    if sys.platform == 'win32':
        return _NOT_WINDOWS
    if sys.platform == 'darwin':
        return _LINUX_ONLY
    return frozenset()


class _LazyGroup(click.Group):
    def __init__(self,
                 *args: Any,
                 lazy_subcommands: dict[str, str] | None = None,
                 **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._lazy_subcommands = lazy_subcommands or {}

    def list_commands(self, ctx: click.Context) -> list[str]:  # noqa: ARG002
        excluded = _excluded()
        return sorted(k for k in self._lazy_subcommands if k not in excluded)

    def get_command(
            self,
            ctx: click.Context,  # noqa: ARG002
            cmd_name: str) -> click.Command | None:
        if cmd_name not in self._lazy_subcommands or cmd_name in _excluded():
            return None
        module_path, func_name = self._lazy_subcommands[cmd_name].rsplit(':', 1)
        cmd: click.Command = getattr(import_module(module_path), func_name)
        cmd.name = cmd_name
        return cmd


@click.group(cls=_LazyGroup, lazy_subcommands=_COMMANDS, context_settings=CONTEXT_SETTINGS)
def main() -> None:
    """Deltona - a collection of utilities."""



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import MagicMock

from deltona.commands.misc import adp_main, burnrariso_main, gogextract_main, unpack_0day_main
from deltona.io import SFVVerificationError, UnRARExtractionTestFailed
import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from click.testing import CliRunner
    from pytest_mock import MockerFixture


@pytest.fixture
def fake_path(tmp_path: Path) -> Path:
    d = tmp_path / 'dir'
    d.mkdir()
    return d


def test_adp_main(runner: CliRunner, mocker: MockerFixture) -> None:
    mock_calc = mocker.patch('deltona.commands.misc.calculate_salary', return_value=12345)
    result = runner.invoke(adp_main, ['--hours', '100', '--pay-rate', '50', '--state', 'FL'])
    assert result.exit_code == 0
    assert '12345' in result.output
    mock_calc.assert_called_once_with(hours=100, pay_rate=50.0, state='FL')


def test_unpack_0day_main(runner: CliRunner, mocker: MockerFixture, fake_path: Path) -> None:
    mock_unpack = mocker.patch('deltona.commands.misc.unpack_0day')
    result = runner.invoke(unpack_0day_main, [str(fake_path)])
    assert result.exit_code == 0
    mock_unpack.assert_called_once_with(fake_path)


def test_gogextract_main(runner: CliRunner, mocker: MockerFixture, tmp_path: Path) -> None:
    fake_file = tmp_path / 'archive.gog'
    fake_file.write_text('data')
    mock_extract = mocker.patch('deltona.commands.misc.extract_gog')
    result = runner.invoke(gogextract_main, [str(fake_file), '-o', str(tmp_path)])
    assert result.exit_code == 0
    mock_extract.assert_called_once_with(fake_file, tmp_path)


def test_burnrariso_main_success(runner: CliRunner, mocker: MockerFixture, tmp_path: Path) -> None:
    rar_file = tmp_path / 'test.rar'
    rar_file.write_text('data')
    iso_file = MagicMock()
    iso_file.name = 'test.iso'
    iso_file.size = 123456
    mock_unrar = mocker.patch('deltona.commands.misc.UnRAR')
    unrar_instance = mock_unrar.return_value
    unrar_instance.list_files.return_value = [iso_file]
    unrar_instance.pipe.return_value.__enter__.return_value.stdout = MagicMock()
    unrar_instance.pipe.return_value.__exit__.return_value = False
    unrar_instance.pipe.return_value.__enter__.return_value.wait = MagicMock(return_value=0)
    unrar_instance.pipe.return_value.__enter__.return_value.returncode = 0
    mock_popen = mocker.patch('deltona.commands.misc.sp.Popen')
    popen_instance = mock_popen.return_value.__enter__.return_value
    popen_instance.wait.return_value = 0
    popen_instance.returncode = 0
    mocker.patch('deltona.commands.misc.verify_sfv')
    mocker.patch('deltona.commands.misc.Path.exists', return_value=True)
    result = runner.invoke(burnrariso_main, [str(rar_file), '--no-crc-check'])
    assert result.exit_code == 0
    unrar_instance.list_files.assert_called_once()
    mock_popen.assert_called()


def test_burnrariso_main_failure(runner: CliRunner, mocker: MockerFixture, tmp_path: Path) -> None:
    rar_file = tmp_path / 'test.rar'
    rar_file.write_text('data')
    iso_file = MagicMock()
    iso_file.name = 'test.iso'
    iso_file.size = 123456
    mock_unrar = mocker.patch('deltona.commands.misc.UnRAR')
    unrar_instance = mock_unrar.return_value
    unrar_instance.list_files.return_value = [iso_file]
    unrar_instance.pipe.return_value.__enter__.return_value.stdout = MagicMock()
    unrar_instance.pipe.return_value.__exit__.return_value = False
    unrar_instance.pipe.return_value.__enter__.return_value.wait = MagicMock(return_value=0)
    unrar_instance.pipe.return_value.__enter__.return_value.returncode = 0
    mock_popen = mocker.patch('deltona.commands.misc.sp.Popen')
    popen_instance = mock_popen.return_value.__enter__.return_value
    popen_instance.wait.return_value = 0
    popen_instance.returncode = 1
    mocker.patch('deltona.commands.misc.verify_sfv')
    mocker.patch('deltona.commands.misc.Path.exists', return_value=True)
    result = runner.invoke(burnrariso_main, [str(rar_file), '--no-crc-check'])
    assert result.exit_code != 0
    unrar_instance.list_files.assert_called_once()
    mock_popen.assert_called()


def test_burnrariso_main_no_iso(runner: CliRunner, mocker: MockerFixture, tmp_path: Path) -> None:
    rar_file = tmp_path / 'test.rar'
    rar_file.write_text('data')
    mock_unrar = mocker.patch('deltona.commands.misc.UnRAR')
    unrar_instance = mock_unrar.return_value
    unrar_instance.list_files.return_value = []
    result = runner.invoke(burnrariso_main, [str(rar_file), '--no-crc-check'])
    assert result.exit_code != 0


def test_burnrariso_main_iso_no_size(runner: CliRunner, mocker: MockerFixture,
                                     tmp_path: Path) -> None:
    rar_file = tmp_path / 'test.rar'
    rar_file.write_text('data')
    iso_file = MagicMock()
    iso_file.name = 'test.iso'
    iso_file.size = 0
    mock_unrar = mocker.patch('deltona.commands.misc.UnRAR')
    unrar_instance = mock_unrar.return_value
    unrar_instance.list_files.return_value = [iso_file]
    result = runner.invoke(burnrariso_main, [str(rar_file), '--no-crc-check'])
    assert result.exit_code != 0


def test_burnrariso_main_sfv_fail(runner: CliRunner, mocker: MockerFixture, tmp_path: Path) -> None:
    rar_file = tmp_path / 'test.rar'
    rar_file.write_text('data')
    iso_file = MagicMock()
    iso_file.name = 'test.iso'
    iso_file.size = 123
    mock_unrar = mocker.patch('deltona.commands.misc.UnRAR')
    unrar_instance = mock_unrar.return_value
    unrar_instance.list_files.return_value = [iso_file]
    mocker.patch('deltona.commands.misc.Path.exists', return_value=True)
    mock_verify = mocker.patch('deltona.commands.misc.verify_sfv',
                               side_effect=SFVVerificationError('.', 1, 0))
    result = runner.invoke(burnrariso_main, [str(rar_file)])
    assert result.exit_code != 0
    mock_verify.assert_called()


def test_burnrariso_main_test_extraction_fail(runner: CliRunner, mocker: MockerFixture,
                                              tmp_path: Path) -> None:
    rar_file = tmp_path / 'test.rar'
    rar_file.write_text('data')
    iso_file = MagicMock()
    iso_file.name = 'test.iso'
    iso_file.size = 123
    mock_unrar = mocker.patch('deltona.commands.misc.UnRAR')
    unrar_instance = mock_unrar.return_value
    unrar_instance.list_files.return_value = [iso_file]
    unrar_instance.test_extraction.side_effect = UnRARExtractionTestFailed
    mocker.patch('deltona.commands.misc.Path.exists', return_value=True)
    mocker.patch('deltona.commands.misc.verify_sfv')
    mocker.patch('deltona.commands.misc.sp.Popen')
    result = runner.invoke(burnrariso_main, [str(rar_file), '--test-extraction'])
    assert result.exit_code != 0



"""Git commands."""

from __future__ import annotations

from time import sleep
from typing import TYPE_CHECKING
import getpass
import re
import webbrowser

from bascom import setup_logging
from deltona.constants import CONTEXT_SETTINGS
from deltona.git import (
    convert_git_ssh_url_to_https,
    get_github_default_branch,
    merge_dependabot_pull_requests,
)
import click
import keyring

if TYPE_CHECKING:
    from git import Repo


def _get_git_repo() -> Repo:  # pragma: no cover
    from git import Repo  # noqa: PLC0415

    return Repo(search_parent_directories=True)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('origin_name', metavar='ORIGIN_NAME', default='origin')
@click.option('-b', '--base-url', help='Base URL for enterprise.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-u', '--username', default=getpass.getuser(), help='Username (passed to keyring).')
def git_checkout_default_branch_main(username: str,
                                     base_url: str | None = None,
                                     origin_name: str = 'origin',
                                     *,
                                     debug: bool = False) -> None:
    """
    Checkout to the default branch.

    For repositories whose origin is on GitHub only.

    To set a token, ``keyring set tmu-github-api "${USER}"``. The token must have
    access to the public_repo or repo scope.
    """  # noqa: DOC501
    setup_logging(debug=debug, loggers={'deltona': {}, 'github': {}, 'keyring': {}})
    token = keyring.get_password('tmu-github-api', username)
    if not token:
        click.echo('No token.', err=True)
        raise click.Abort
    repo = _get_git_repo()
    default_branch = get_github_default_branch(repo=repo,
                                               base_url=base_url,
                                               token=token,
                                               origin_name=origin_name)
    next(b for b in repo.heads if b.name == default_branch).checkout()


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('origin_name', metavar='ORIGIN_NAME', default='origin')
@click.option('-b', '--base-url', help='Base URL for enterprise.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-u', '--username', default=getpass.getuser(), help='Username (passed to keyring).')
@click.option('-r',
              '--remote',
              is_flag=True,
              help='Rebase with the origin copy of the default branch.')
def git_rebase_default_branch_main(
    username: str,
    base_url: str | None = None,
    origin_name: str = 'origin',
    *,
    debug: bool = False,
    remote: bool = False,
) -> None:
    """
    Rebase the current head with the default branch.

    For repositories whose origin is on GitHub only.

    To set a token, ``keyring set tmu-github-api "${USER}"``. The token must have
    access to the public_repo or repo scope.
    """  # noqa: DOC501
    setup_logging(debug=debug, loggers={'deltona': {}, 'github': {}, 'keyring': {}})
    token = keyring.get_password('tmu-github-api', username)
    if not token:
        click.echo('No token.', err=True)
        raise click.Abort
    repo = _get_git_repo()
    default_branch = get_github_default_branch(repo=repo,
                                               base_url=base_url,
                                               token=token,
                                               origin_name=origin_name)
    repo.git.rebase(f'{origin_name}/{default_branch}' if remote else default_branch)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('name', default='origin')
def git_open_main(name: str = 'origin') -> None:
    """Open assumed repository web representation (GitHub, GitLab, etc) based on the origin."""
    url = _get_git_repo().remote(name).url
    if re.search(r'^https?://', url):
        webbrowser.open(url)
        return
    webbrowser.open(convert_git_ssh_url_to_https(url))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    '-a',
    '--affiliation',
    default='owner',
    help='Affiliation. See REST API documentation for more information.',
)
@click.option('-b', '--base-url', help='Base URL for enterprise.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('--delay', type=float, default=120, help='Delay in seconds between attempts.')
@click.option('-u', '--username', default=getpass.getuser(), help='Username.')
def merge_dependabot_prs_main(
    username: str,
    affiliation: str = 'owner',
    base_url: str | None = None,
    delay: float = 120,
    *,
    debug: bool = False,
) -> None:
    """Merge pull requests made by Dependabot on GitHub."""  # noqa: DOC501
    setup_logging(debug=debug, loggers={'deltona': {}, 'github': {}, 'keyring': {}})
    if not (token := keyring.get_password('tmu-github-api', username)):
        click.echo('No token.', err=True)
        raise click.Abort
    while True:
        try:
            merge_dependabot_pull_requests(affiliation=affiliation, base_url=base_url, token=token)
            break
        except RuntimeError:
            click.echo(f'Sleeping for {delay} seconds.')
            sleep(delay)



"""UltraISO wrapper."""

from __future__ import annotations

from pathlib import Path
from shlex import quote
from shutil import copyfile
from typing import TYPE_CHECKING, Literal
import logging
import os
import subprocess as sp

from typing_extensions import override

from .string import unix_path_to_wine as base_unix_path_to_wine
from .system import IS_WINDOWS

if TYPE_CHECKING:
    from collections.abc import Iterable

    from .typing import StrPath, StrPathMustExist

__all__ = ('InvalidExec', 'patch_ultraiso_font', 'run_ultraiso')

DEFAULT_WINE_PREFIX = Path.home() / '.local/share/wineprefixes/ultraiso'
MIN_ARGUMENTS = 4 if not IS_WINDOWS else 3
log = logging.getLogger(__name__)


def unix_path_to_wine(s: StrPath) -> str:
    if IS_WINDOWS:
        return str(s)
    return base_unix_path_to_wine(str(s))


def get_ultraiso_path(prefix: StrPath) -> StrPath | None:
    program_files_paths = [f'Program Files{x}/UltraISO/UltraISO.exe' for x in ('', ' (x86)')]
    prefix = Path(prefix) / 'drive_c' if not IS_WINDOWS else Path('C:/')
    for exe in (prefix / x for x in program_files_paths):
        if exe.exists():
            return exe
    return None


class InsufficientArguments(Exception):
    @override
    def __init__(self) -> None:
        super().__init__('Insufficient amount of arguments.')


def run_ultraiso(  # noqa: PLR0913
    *,
    add_dirs: Iterable[StrPathMustExist] | None = None,
    add_files: Iterable[StrPathMustExist] | None = None,
    cmd: StrPathMustExist | None = None,
    input: StrPathMustExist | None = None,  # noqa: A002
    output: StrPath | None = None,
    appid: str | None = None,
    preparer: str | None = None,
    publisher: str | None = None,
    sysid: str | None = None,
    volset: int | None = None,
    volume: str | None = None,
    ilong: bool = False,
    imax: bool = False,
    lowercase: bool = False,
    vernum: bool = False,
    hfs: bool = False,
    jlong: bool = False,
    joliet: bool = False,
    rockridge: bool = False,
    udf: bool = False,
    udfdvd: bool = False,
    bootfile: StrPathMustExist | None = None,
    bootinfotable: bool = False,
    optimize: bool = False,
    chdir: str | None = None,
    newdir: str | None = None,
    rmdir: str | None = None,
    ahide: str | None = None,
    hide: str | None = None,
    pn: Literal[1, 2, 3, 4, 5, 6, 7, 8, 9] | None = None,
    bin2iso: StrPathMustExist | None = None,
    dmg2iso: StrPathMustExist | None = None,
    bin2isz: StrPathMustExist | None = None,
    compress: Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16] | None = None,
    encrypt: Literal[1, 2, 3] | None = None,
    password: str | None = None,
    split: int | None = None,
    extract: StrPath | None = None,
    get: str | None = None,
    list_: StrPath | None = None,
    prefix: StrPathMustExist = DEFAULT_WINE_PREFIX,
) -> None:
    r"""
    Run UltraISO in a convenient way.

    Despite always passing ``-silent`` to the application, windows including ones requiring
    interaction may still appear.

    On non-Windows, for any option taking a file path, only pass UNIX file paths. They will be
    converted as needed.

    Parameters
    ----------
    input : StrPathMustExist | None
        Input image.
    output : StrPath | None
        Output image.
    add_dirs : Iterable[StrPath] | None
        Directories to add.
    add_files : Iterable[StrPath] | None
        Files to add.
    cmd : StrPathMustExist | None
        File to read command line arguments from. If this argument is present and not ``None``, all
        other arguments are ignored.
    appid : str | None
        Application ID.
    preparer : str | None
        Preparer name.
    publisher : str | None
        Publisher name.
    sysid : str | None
        System ID.
    volset : int | None
        Volume set ID.
    volume : str | None
        Volume label.
    ilong : bool
        Long filename for ISO 9660 volume, up to 31 characters.
    imax : bool
        Max filename for ISO 9660 volume, up to 207 characters.
    lowercase : bool
        Allow lowercase letters.
    vernum : bool
        Include file version number.
    hfs : bool
        Create Apple HFS volume.
    jlong : bool
        Long filenames for joliet volume, up to 103 characters.
    joliet : bool
        Create a Joliet volume.
    rockridge : bool
        Create RockRidge volume.
    udf : bool
        Create UDF volume.
    udfdvd : bool
        Create UDF DVD image (this option will overwrite all other volume settings).
    bootfile : StrPathMustExist | None
        Boot file.
    bootinfotable : bool
        Generate boot information table in boot file.
    optimize : bool
        Optimise file systems by coding same files only once.
    chdir : str | None
        Change current directory in ISO image.
    newdir : str | None
        Create a new directory.
    rmdir : str | None
        Remove a file or folder from ISO image (full path should be specified).
    ahide : str | None
        Set advanced hidden attribute of a file or folder (full path should be specified).
    hide : str | None
        Set hidden attribute of a file or folder (full path should be specified).
    pn : Literal[1, 2, 3, 4, 5, 6, 7, 8, 9] | None
        Set priority of a file or folder (full path should be specified).
    bin2iso : StrPathMustExist | None
        Convert input CD/DVD image to ISO format.
    dmg2iso : StrPathMustExist | None
        Convert input DMG image to ISO format.
    bin2isz : StrPathMustExist | None
        Convert image to ISZ.
    compress : Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16] | None
        Set compression level.
    encrypt : Literal[1, 2, 3] | None
        Set encryption method.
    password : str | None
        Set password.
    split : int | None
        Set segment size in bytes.
    extract : StrPath | None
        Extract ISO image to specified directory.
    get : str | None
        Set a file or directory (full path) to be extracted.
    list\_ : StrPath | None
        Create a list of files and directories in an ISO image.
    prefix : StrPathMustExist
        Wine prefix path.

    Raises
    ------
    InsufficientArguments
        If not enough arguments are passed.
    FileNotFoundError
        If the UltraISO executable cannot be found.
    """  # noqa: DOC501
    if (actual_exe_path := get_ultraiso_path(prefix)) is None:
        raise FileNotFoundError
    env = {}
    if not IS_WINDOWS:
        env = {'WINEPREFIX': str(prefix), 'HOME': os.environ['HOME']}
        if 'DISPLAY' not in os.environ or 'XAUTHORITY' not in os.environ:  # pragma: no cover
            log.warning(
                'UltraISO.exe will likely fail to run since DISPLAY or XAUTHORITY are not in the '
                'environment.')
        env['DISPLAY'] = os.environ.get('DISPLAY', '')
        env['XAUTHORITY'] = os.environ.get('XAUTHORITY', '')
        env['WINEDEBUG'] = 'fixme-all'
    sp_args: list[str] = ['wine'] if not IS_WINDOWS else []
    sp_args += [str(actual_exe_path), '-silent']
    if cmd:
        sp_args += ['-cmd', unix_path_to_wine(cmd)]
    else:
        for key, filename in (('-in', input), ('-out', output)):
            if filename:
                sp_args += [key, unix_path_to_wine(filename)]
        for file in add_files or []:
            sp_args += ['-file', unix_path_to_wine(file)]
        for dir_ in add_dirs or []:
            sp_args += ['-directory', str(dir_)]
        for k in (k for k, v in {
                'bootinfotable': bootinfotable,
                'hfs': hfs,
                'ilong': ilong,
                'imax': imax,
                'jlong': jlong,
                'joliet': joliet,
                'lowercase': lowercase,
                'optimize': optimize,
                'rockridge': rockridge,
                'udf': udf,
                'udfdvd': udfdvd,
                'vernum': vernum,
        }.items() if v):
            sp_args += [f'-{k}']
        for k, v in ((k, v) for k, v in {
                'bootfile': bootfile,
                'bin2iso': bin2iso,
                'dmg2iso': dmg2iso,
                'bin2isz': bin2isz,
        }.items() if v is not None):
            sp_args += [f'-{k}', unix_path_to_wine(v)]
        for k, v in ((k, v) for k, v in {
                'appid': appid,
                'preparer': preparer,
                'publisher': publisher,
                'sysid': sysid,
                'volume': volume,
                'chdir': chdir,
                'newdir': newdir,
                'rmdir': rmdir,
                'ahide': ahide,
                'hide': hide,
                'password': password,
                'extract': extract,
                'get': get,
                'list': list_,
        }.items() if v is not None):
            sp_args += [f'-{k}', str(v)]
        for k, i in ((k, i) for k, i in {
                'volset': volset,
                'compress': compress,
                'encrypt': encrypt,
                'split': split,
                'pn': pn,
        }.items() if i is not None):
            sp_args += [f'-{k}', str(i)]
    if len(sp_args) < MIN_ARGUMENTS:
        raise InsufficientArguments
    quoted_args = ' '.join(quote(x) for x in sp_args)
    if not IS_WINDOWS:
        log.debug('Command: env %s %s', quote(f'WINEPREFIX={prefix}'), quoted_args)
    else:
        log.debug('Command: %s', quoted_args)
    try:
        sp.run(sp_args, env=env, check=True)
    except sp.CalledProcessError as e:
        if stderr := e.stderr.strip():
            log.exception('stderr output:')
            for line in stderr.splitlines():
                if not IS_WINDOWS and ('winemenubuilder.exe' in line or 'fixme:' in line
                                       or 'wine: using fast synchronization.' in line):
                    continue
                log.exception(' -> %s', line)
        raise


ULTRAISO_FONT_REPLACEMENT_MAX_LENGTH = 13


class InvalidExec(Exception):
    """Raised when the UltraISO executable does not contain the expected font string."""
    @override
    def __init__(self) -> None:
        super().__init__('Font not found in file. This is not the original UltraISO executable.')


def patch_ultraiso_font(exe: os.PathLike[str], font_name: str = 'Noto Sans') -> None:
    """
    Patch the default font in the UltraISO executable.

    Replaces ``MS Sans Serif`` with the specified font name. A backup of the original file is
    created with a ``.exebak`` extension.

    Parameters
    ----------
    exe : os.PathLike[str]
        Path to the UltraISO executable.
    font_name : str
        Replacement font name. Maximum 13 characters.

    Raises
    ------
    ValueError
        If ``font_name`` exceeds the maximum length.
    FileNotFoundError
        If ``exe`` does not exist.
    IsADirectoryError
        If ``exe`` or the backup path is a directory.
    InvalidExec
        If the executable does not contain the expected font string.
    """
    exe = Path(exe)
    if len(font_name) > ULTRAISO_FONT_REPLACEMENT_MAX_LENGTH:
        msg = f'Font name too long. Max length is {ULTRAISO_FONT_REPLACEMENT_MAX_LENGTH}.'
        raise ValueError(msg)
    if not exe.exists():
        msg = f'File {exe} does not exist.'
        raise FileNotFoundError(msg)
    if not exe.is_file():
        msg = f'File {exe} is a directory.'
        raise IsADirectoryError(msg)
    backup_file = exe.with_suffix('.exebak')
    if backup_file.exists():
        if not backup_file.is_file():
            msg = f'File {backup_file} is a directory.'
            raise IsADirectoryError(msg)
    else:
        copyfile(exe, exe.with_suffix('.exebak'))
    data = backup_file.read_bytes()
    if b'MS Sans Serif\x00' not in data:
        raise InvalidExec
    exe.write_bytes(data.replace(b'MS Sans Serif\x00', font_name.encode() + b'\x00'))



"""Gentoo system utilities."""

from __future__ import annotations

from fnmatch import fnmatch
from pathlib import Path
from shutil import rmtree
from typing import TYPE_CHECKING
import logging
import re

from typing_extensions import override

if TYPE_CHECKING:
    from collections.abc import Iterator

    from .typing import StrPath

__all__ = (
    'DEFAULT_ACTIVE_KERNEL_NAME',
    'DEFAULT_KERNEL_LOCATION',
    'DEFAULT_MODULES_PATH',
    'InvalidActiveKernelSourcePath',
    'clean_old_kernels_and_modules',
)

DEFAULT_ACTIVE_KERNEL_NAME = 'linux'
"""
Default kernel symlink name.

:meta hide-value:
"""
DEFAULT_KERNEL_LOCATION = Path('/usr/src')
"""
Default kernel sources location.

:meta hide-value:
"""
DEFAULT_MODULES_PATH = Path('/lib/modules/')
"""
Default kernel modules location.

:meta hide-value:
"""
log = logging.getLogger(__name__)


class InvalidActiveKernelSourcePath(Exception):
    """Raised when the active kernel source path is not a symbolic link."""
    @override
    def __init__(self, source_path: StrPath) -> None:
        super().__init__(f'{source_path} is not a symbolic link.')


def clean_old_kernels_and_modules(
    path: StrPath = DEFAULT_KERNEL_LOCATION,
    modules_path: StrPath = DEFAULT_MODULES_PATH,
    active_kernel_name: str = DEFAULT_ACTIVE_KERNEL_NAME,
) -> Iterator[str]:
    """
    Remove inactive kernels and modules.

    By default, removes old Linux sources from ``/usr/src``.

    Parameters
    ----------
    path : StrPath
        Location where kernel sources are installed. Defaults to ``/usr/src``.
    modules_path : StrPath
        Location where modules get installed, such as ``/lib/modules``.
    active_kernel_name : str
        Kernel name like ``'linux'``.

    Yields
    ------
    str
        Deleted path.

    Raises
    ------
    InvalidActiveKernelSourcePath
        If the active kernel source path is not a symbolic link.
    FileNotFoundError
        If the active kernel source path does not exist.
    """
    path = Path(path)
    modules_path = Path(modules_path)
    loc = path / active_kernel_name
    if not loc.exists():
        raise FileNotFoundError
    if not loc.is_symlink():
        raise InvalidActiveKernelSourcePath(loc)
    v = re.sub(rf'^{DEFAULT_ACTIVE_KERNEL_NAME}-', '', str(loc.readlink()))
    current_version = f'{active_kernel_name}-{v}'
    version_pat = f'*{v}*'
    for d in modules_path.glob('*'):
        log.debug('Examining: %s', d)
        if d.is_dir() and not fnmatch(str(d), version_pat):
            log.debug('Will delete: %s', d)
            rmtree(d)
            yield str(d)
    for d in path.glob(f'{active_kernel_name}-*'):
        log.debug('Examining: %s', d)
        if d.is_dir() and d.name != current_version:
            log.debug('Will delete: %s', d)
            rmtree(d)
            yield str(d)
